import logging
from abc import abstractmethod

from easy_tpp.preprocess import TPPDataLoader
from easy_tpp.utils import Registrable, Timer, logger, get_unique_id, LogConst, get_stage, RunnerPhase


class Runner(Registrable):
    """Registrable Base Runner class.
    """

    def __init__(
            self,
            runner_config,
            unique_model_dir=False,
            **kwargs):
        """Initialize the base runner.

        Args:
            runner_config (RunnerConfig): config for the runner.
            unique_model_dir (bool, optional): whether to give unique dir to save the model. Defaults to False.
        """
        self.runner_config = runner_config
        # re-assign the model_dir
        if unique_model_dir:
            runner_config.model_dir = runner_config.base_config.specs['saved_model_dir'] + '_' + get_unique_id()

        self.save_log()

        skip_data_loader = kwargs.get('skip_data_loader', False)
        if not skip_data_loader:
            # build data reader
            data_config = self.runner_config.data_config
            backend = self.runner_config.base_config.backend
            kwargs = self.runner_config.trainer_config.get_yaml_config()
            self._data_loader = TPPDataLoader(
                data_config=data_config,
                backend=backend,
                **kwargs
            )
        self.timer = Timer()

    @staticmethod
    def build_from_config(runner_config, unique_model_dir=False, **kwargs):
        """Build up the runner from runner config.

        Args:
            runner_config (RunnerConfig): config for the runner.
            unique_model_dir (bool, optional): whether to give unique dir to save the model. Defaults to False.

        Returns:
            Runner: the corresponding runner class.
        """
        runner_cls = Runner.by_name(runner_config.base_config.runner_id)
        return runner_cls(runner_config, unique_model_dir=unique_model_dir, **kwargs)

    def get_config(self):
        return self.runner_config

    def set_model_dir(self, model_dir):
        self.runner_config.base_config.specs['saved_model_dir'] = model_dir

    def get_model_dir(self):
        return self.runner_config.base_config.specs['saved_model_dir']

    def train(
            self,
            train_loader=None,
            valid_loader=None,
            test_loader=None,
            **kwargs
    ):
        """Train the model.

        Args:
            train_loader (EasyTPP.DataLoader, optional): data loader for train set. Defaults to None.
            valid_loader (EasyTPP.DataLoader, optional): data loader for valid set. Defaults to None.
            test_loader (EasyTPP.DataLoader, optional): data loader for test set. Defaults to None.

        Returns:
            model: _description_
        """
        # no train and valid loader from outside
        if train_loader is None and valid_loader is None:
            train_loader = self._data_loader.train_loader()
            valid_loader = self._data_loader.valid_loader()

        # no test loader from outside and there indeed exits test data in config
        if test_loader is None and self.runner_config.data_config.test_dir is not None:
            test_loader = self._data_loader.test_loader()

        logger.info(f'Data \'{self.runner_config.base_config.dataset_id}\' loaded...')

        timer = self.timer
        timer.start()
        model_id = self.runner_config.base_config.model_id
        logger.info(f'Start {model_id} training...')
        model = self._train_model(
            train_loader,
            valid_loader,
            test_loader=test_loader,
            **kwargs
        )
        logger.info(f'End {model_id} train! Cost time: {timer.end()}')
        return model

    def evaluate(self, valid_loader=None, **kwargs):
        if valid_loader is None:
            valid_loader = self._data_loader.valid_loader()

        logger.info(f'Data \'{self.runner_config.base_config.dataset_id}\' loaded...')

        timer = self.timer
        timer.start()
        model_id = self.runner_config.base_config.model_id
        logger.info(f'Start {model_id} evaluation...')

        metric = self._evaluate_model(
            valid_loader,
            **kwargs
        )
        logger.info(f'End {model_id} evaluation! Cost time: {timer.end()}')
        return metric['rmse']  # return a list of scalr for HPO to use

    def gen(self, gen_loader=None, **kwargs):
        if gen_loader is None:
            gen_loader = self._data_loader.test_loader()

        logger.info(f'Data \'{self.runner_config.base_config.dataset_id}\' loaded...')

        timer = self.timer
        timer.start()
        model_name = self.runner_config.base_config.model_id
        logger.info(f'Start {model_name} evaluation...')

        model = self._gen_model(
            gen_loader,
            **kwargs
        )
        logger.info(f'End {model_name} generation! Cost time: {timer.end()}')
        return model

    @abstractmethod
    def _train_model(self, train_loader, valid_loader, **kwargs):
        pass

    @abstractmethod
    def _evaluate_model(self, data_loader, **kwargs):
        pass

    @abstractmethod
    def _gen_model(self, data_loader, **kwargs):
        pass

    @abstractmethod
    def _save_model(self, model_dir, **kwargs):
        pass

    @abstractmethod
    def _load_model(self, model_dir, **kwargs):
        pass

    def save_log(self):
        """Save log to local files
        """
        log_dir = self.runner_config.base_config.specs['saved_log_dir']
        fh = logging.FileHandler(log_dir)
        fh.setFormatter(logging.Formatter(LogConst.DEFAULT_FORMAT_LONG))
        logger.addHandler(fh)
        logger.info(f'Save the log to {log_dir}')
        return

    def save(
            self,
            model_dir=None,
            **kwargs
    ):
        return self._save_model(model_dir, **kwargs)

    def run(self, **kwargs):
        """Start the runner.

        Args:
            **kwargs (dict): optional params.

        Returns:
            EasyTPP.BaseModel, dict: the results of the process.
        """
        current_stage = get_stage(self.runner_config.base_config.stage)
        if current_stage == RunnerPhase.TRAIN:
            return self.train(**kwargs)
        elif current_stage == RunnerPhase.VALIDATE:
            return self.evaluate(**kwargs)
        else:
            return self.gen(**kwargs)



from easy_tpp.runner.base_runner import Runner
from easy_tpp.runner.tpp_runner import TPPRunner
# for register all necessary contents
from easy_tpp.default_registers.register_metrics import *

__all__ = ['Runner',
           'TPPRunner']


import os
import time


def is_master_process():
    """ Check if the process is the master process in all machines.

    Returns:
        bool
    """
    rank = 0 if os.getenv('RANK') is None else int(os.getenv('RANK'))
    if rank == 0:
        return True
    else:
        return False


def is_local_master_process():
    """ Check if the process is the master process in the local machine.

    Returns:
        bool
    """
    rank = 0 if os.getenv('RANK') is None else int(os.getenv('RANK'))
    local_world_size = 1 if os.getenv('LOCAL_WORLD_SIZE') is None else int(os.getenv('LOCAL_WORLD_SIZE'))
    if local_world_size == 0 or rank % local_world_size == 0:
        return True
    else:
        return False


def get_now_timestamp_id():
    """ Get the current timestamp string.

    Returns:
        A string like yyMMdd_hhmmss
    """
    import datetime
    return datetime.datetime.now().strftime('%y%m%d-%H%M%S')


def get_unique_id():
    """ Generate a unique id string based on process id (pid), thread id and timestamp.

    Returns:
        Unique id: str
    """
    import os
    import threading
    pid = os.getpid()
    tid = threading.currentThread().ident
    ts_id = get_now_timestamp_id()

    return '{}_{}_{}'.format(pid, tid, ts_id)


def parse_uri_to_protocol_and_path(uri):
    """ Parse a uri into two parts, protocol and path. Set 'file' as default protocol when lack protocol.

    Args:
        uri: str
            The uri to identify a resource, whose format is like 'protocol://uri'.

    Returns:
        Protocol: str. The method to access the resource.
        URI: str. The location of the resource.
    """

    if uri is None:
        return None, None
    tokens = uri.split('://')
    if len(tokens) == 2:
        protocol = tokens[0]
        path = tokens[1]
    elif len(tokens) == 1:
        protocol = 'file'
        path = tokens[0]
    else:
        raise RuntimeError(f'Wrong url format: {uri}')

    return protocol, path


class Timer:
    """Count the elapsing time between start and end.
    """

    def __init__(self, unit='m'):
        unit = unit.lower()
        if unit == 's':
            self._unit = 1
        elif unit == 'm':
            self._unit = 60
        elif unit == 'h':
            self._unit = 1440
        else:
            raise RuntimeError('Unknown unit:', unit)

        self.unit = unit
        # default start time is set to the time the object initialized
        self._start_time = time.time()

    def start(self):
        self._start_time = time.time()

    def end(self):
        end_time = time.time()
        cost = (end_time - self._start_time) / self._unit
        # reset the start time using the end time
        self._start_time = end_time
        return '%.3f%s' % (cost, self.unit)


# -------------------------- Singleton Object --------------------------
default_timer = Timer()



import logging
import sys
import typing

from easy_tpp.utils.const import LogConst

# -------- log setting ---------
DEFAULT_LOGGER = "easytpp.logger"


class CustomFormatter(logging.Formatter):
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = LogConst.DEFAULT_FORMAT_LONG

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


DEFAULT_FORMATTER = CustomFormatter()

_ch = logging.StreamHandler(stream=sys.stdout)
_ch.setFormatter(DEFAULT_FORMATTER)

_DEFAULT_HANDLERS = [_ch]

_LOGGER_CACHE = {}  # type: typing.Dict[str, logging.Logger]


def get_logger(name, level="INFO", handlers=None, update=False):
    if name in _LOGGER_CACHE and not update:
        return _LOGGER_CACHE[name]
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.handlers = handlers or _DEFAULT_HANDLERS
    logger.propagate = False
    return logger


# -------------------------- Singleton Object --------------------------
default_logger = get_logger(DEFAULT_LOGGER)



import math

import numpy as np
import tensorflow as tf
from tensorflow.python.keras import layers

from easy_tpp.model.tf_model.tf_baselayer import EncoderLayer
from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel
from easy_tpp.utils.tf_utils import get_shape_list

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


class AttNHP(TfBaseModel):
    def __init__(self, model_config):
        super(AttNHP, self).__init__(model_config)
        self.d_model = model_config.hidden_size
        self.d_time = model_config.data_specs['time_emb_size']
        self.use_norm = model_config.data_specs['use_ln']

        self.n_layers = model_config.data_specs['num_layers']
        self.n_head = model_config.data_specs['num_heads']
        self.dropout = model_config.dropout_rate

        # position vector, used for temporal encoding
        self.div_term = self.make_div_term()

    def make_div_term(self):
        """Initialize the division term used in temporal embedding.

        Returns:
            np.array: diviser.
        """
        div_term_ = np.exp(np.arange(0, self.d_time, 2) * -(math.log(10000.0) / self.d_time))
        div_term_ = np.reshape(div_term_, (1, 1, -1))

        div_term = np.ones((1, 1, self.d_time))

        for i in range(self.d_time):
            div_term[..., i] = div_term_[..., i // 2]

        return div_term

    def build_graph(self):
        """Build up the network
        """
        with tf.variable_scope('AttNHP'):
            self.build_input_graph()

            self.layer_intensity = layers.Dense(self.num_event_types, activation=tf.nn.softplus)

            self.heads = []
            for i in range(self.n_head):
                self.heads.append([EncoderLayer(hidden_size=self.d_model,
                                                num_heads=self.n_head,
                                                dropout_rate=self.dropout) for _ in range(self.n_layers)])

            self.loss, self.num_event = self.loglike_loss()

            # Make predictions
            if self.event_sampler and self.gen_config.num_step_gen == 1:
                self.dtime_predict_one_step, self.type_predict_one_step = \
                    self.predict_one_step_at_every_event(self.time_seqs,
                                                         self.time_delta_seqs,
                                                         self.type_seqs)

            if self.event_sampler and self.gen_config.num_step_gen > 1:
                # make generations
                self.dtime_generation, self.type_generation = \
                    self.predict_multi_step_since_last_event(self.time_seqs,
                                                             self.time_delta_seqs,
                                                             self.type_seqs,
                                                             num_step=self.gen_config.num_step_gen)

    def compute_temporal_embedding(self, time_seqs):
        """Compute the temporal embedding.

        Args:
            time_seqs (tensor): [batch_size, seq_len].

        Returns:
            tensor: [batch_size, seq_len, emb_size].
        """

        batch_size, seq_len = get_shape_list(time_seqs)  # dynamic

        # [self.d_time]
        position_mask = np.array([1] * self.d_time)
        position_mask[1::2] = 0  # dim 2i+1
        position_mask = tf.convert_to_tensor(position_mask, tf.int32)

        # [batch_size, max_len, d_time]
        position_mask = tf.tile(position_mask[None, None, ...], [batch_size, seq_len, 1])

        time_seqs_ = time_seqs[..., None]

        position_enc = tf.where(tf.equal(position_mask, 0),  # dim 2i+1
                                tf.cos(time_seqs_ * self.div_term),  # dim 2i+1
                                tf.sin(time_seqs_ * self.div_term))  # dim 2i

        # [batch_size, max_len, hidden_dim]
        return position_enc

    def seq_encoding(self, time_seqs, type_seqs):
        """Encode the sequence.

        Args:
            time_seqs (tensor): time seqs input, [batch_size, seq_len].
            event_seqs (_type_): event type seqs input, [batch_size, seq_len].

        Returns:
            tuple: event embedding, time embedding and type embedding.
        """

        # [batch_size, seq_len, hidden_size]
        time_emb = self.compute_temporal_embedding(time_seqs)
        # [batch_size, seq_len, hidden_size]
        type_emb = tf.tanh(self.layer_type_emb(type_seqs))
        # [batch_size, seq_len, hidden_size*2]
        event_emb = tf.concat([type_emb, time_emb], axis=-1)

        return event_emb, time_emb, type_emb

    def make_layer_mask(self, attention_mask):
        """Create a tensor to do masking on layers.

        Args:
            attention_mask (tensor): mask for attention operation, [batch_size, seq_len, seq_len]

        Returns:
            tensor: aim to keep the current layer, the same size of attention mask
            a diagonal matrix, [batch_size, seq_len, seq_len]
        """
        # [batch_size, seq_len, seq_len]
        layer_mask = tf.eye(get_shape_list(attention_mask)[1]) < 1
        layer_mask = layer_mask[None, ...]
        layer_mask = tf.tile(layer_mask, [get_shape_list(attention_mask)[0], 1, 1])
        return tf.cast(layer_mask, tf.int32)

    def make_combined_att_mask(self, attention_mask, layer_mask):
        """Combined attention mask and layer mask.

        Args:
            attention_mask (tensor): mask for attention operation, [batch_size, seq_len, seq_len]
            layer_mask (tensor): mask for other layers, [batch_size, seq_len, seq_len]

        Returns:
            tensor: [batch_size, seq_len * 2, seq_len * 2]
        """

        # [batch_size, seq_len, seq_len * 2]
        combined_mask = tf.concat([attention_mask, layer_mask], axis=-1)
        # [batch_size, seq_len, seq_len * 2]
        contextual_mask = tf.concat([attention_mask, tf.ones_like(layer_mask)], axis=-1)
        # [batch_size, seq_len * 2, seq_len * 2]
        combined_mask = tf.concat([contextual_mask, combined_mask], axis=1)
        return combined_mask

    def forward_pass(self, init_cur_layer, time_emb, sample_time_emb, event_emb, combined_mask):
        """update the structure sequentially.

        Args:
            init_cur_layer (tensor): [batch_size, seq_len, hidden_size]
            time_emb (tensor): [batch_size, seq_len, hidden_size]
            sample_time_emb (tensor): [batch_size, seq_len, hidden_size]
            event_emb (tensor): [batch_size, seq_len, hidden_size]
            combined_mask (tensor): [batch_size, seq_len, hidden_size]

        Returns:
            tensor: [batch_size, seq_len, hidden_size*2]
        """
        cur_layers = []
        seq_len = get_shape_list(time_emb)[1]

        for head_i in range(self.n_head):
            # [batch_size, seq_len, hidden_size]
            cur_layer_ = init_cur_layer
            for layer_i in range(self.n_layers):
                # each layer concats the temporal emb
                # [batch_size, seq_len, hidden_size*2]
                layer_ = tf.concat([cur_layer_, sample_time_emb], axis=-1)
                # make combined input from event emb + layer emb
                # [batch_size, seq_len*2, hidden_size*2]
                _combined_input = tf.concat([event_emb, layer_], axis=1)

                enc_layer = self.heads[head_i][layer_i]
                # compute the output
                enc_output = enc_layer((_combined_input, combined_mask),
                                       training=self.is_training)

                # the layer output
                # [batch_size, seq_len, hidden_size]
                _cur_layer_ = enc_output[:, seq_len:, :]
                # add residual connection
                cur_layer_ = tf.tanh(_cur_layer_) + cur_layer_

                # event emb
                event_emb = tf.concat([enc_output[:, :seq_len, :], time_emb], axis=-1)

                if self.use_norm:
                    cur_layer_ = self.norm(cur_layer_)
            cur_layers.append(cur_layer_)
        cur_layer_ = tf.concat(cur_layers, axis=-1)

        return cur_layer_

    def forward(self, time_seqs, type_seqs, attention_mask, sample_times=None):
        """ Move forward through the network """

        # [batch_size, seq_len, hidden_size]
        event_emb, time_emb, type_emb = self.seq_encoding(time_seqs, type_seqs)
        init_cur_layer = tf.zeros_like(type_emb)
        layer_mask = self.make_layer_mask(attention_mask)

        if sample_times is None:
            sample_time_emb = time_emb
        else:
            sample_time_emb = self.compute_temporal_embedding(sample_times)
        combined_mask = self.make_combined_att_mask(attention_mask, layer_mask)

        encoder_output = self.forward_pass(init_cur_layer, time_emb, sample_time_emb, event_emb, combined_mask)

        return encoder_output

    def loglike_loss(self):
        # 1. compute event-loglik
        enc_out = self.forward(self.time_seqs[:, :-1],
                               self.type_seqs[:, :-1],
                               self.attention_mask[:, 1:, :-1],
                               self.time_seqs[:, 1:])

        # [batch_size, seq_len, num_event_types]
        lambda_at_event = self.layer_intensity(enc_out)

        # 2. compute non-event-loglik (using MC sampling to compute integral)
        # 2.1 sample times
        # [batch_size, seq_len, num_sample]
        temp_time = self.make_dtime_loss_samples(self.time_delta_seqs[:, 1:])

        # [batch_size, seq_len, num_sample]
        sample_times = temp_time + self.time_seqs[:, :-1][..., None]

        # 2.2 compute intensities at sampled times
        # [batch_size, num_times = max_len - 1, num_sample, event_num]
        lambda_t_sample = self.compute_intensities_at_sample_times(self.time_seqs[:, :-1],
                                                                   self.time_delta_seqs[:, :-1],  # not used
                                                                   self.type_seqs[:, :-1],
                                                                   sample_times,
                                                                   attention_mask=self.attention_mask[:, 1:, :-1])

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=self.time_delta_seqs[:, 1:],
                                                                        seq_mask=self.batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=self.type_mask[:, 1:])

        # return enc_inten to compute accuracy
        loss = - tf.reduce_sum(event_ll - non_event_ll)

        return loss, num_events

    def compute_states_at_sample_times(self,
                                       time_seqs,
                                       type_seqs,
                                       attention_mask,
                                       sample_times):
        """

        Args:
            time_seqs: [batch_size, seq_len]
            type_seqs: [batch_size, seq_len]
            attention_mask: [batch_size, seq_len, seq_len]
            sample_times: [batch_size, seq_len, num_samples]

        Returns:
            hidden states at all sampled times: [batch_size, seq_len, num_samples, hidden_size]

        """
        batch_size, seq_len = get_shape_list(type_seqs)
        num_samples = get_shape_list(sample_times)[-1]

        # [num_samples, batch_size, seq_len]
        sample_times = tf.transpose(sample_times, perm=(2, 0, 1))
        # [num_samples * batch_size, seq_len]
        _sample_time = tf.reshape(sample_times, (num_samples * batch_size, -1))
        # [num_samples * batch_size, seq_len]
        _types = tf.reshape(tf.tile(type_seqs[None, ...], (num_samples, 1, 1)), (num_samples * batch_size, -1))
        # [num_samples * batch_size, seq_len]
        _times = tf.reshape(tf.tile(time_seqs[None, ...], (num_samples, 1, 1)), (num_samples * batch_size, -1))
        # [num_samples * batch_size, seq_len]
        _attn_mask = tf.tile(attention_mask[None, ...], (num_samples, 1, 1, 1))
        _attn_mask = tf.reshape(_attn_mask, (num_samples * batch_size, seq_len, seq_len))

        # [num_samples * batch_size, seq_len, hidden_size]
        encoder_output = self.forward(_times,
                                      _types,
                                      _attn_mask,
                                      _sample_time)

        # [num_samples, batch_size, seq_len, hidden_size]
        encoder_output = tf.reshape(encoder_output, (num_samples, batch_size, seq_len, -1))
        # [batch_size, seq_len, num_samples, hidden_size]
        encoder_output = tf.transpose(encoder_output, perm=(1, 2, 0, 3))
        return encoder_output

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sample_times, **kwargs):
        """
        Args:
            time_seqs: [batch_size, seq_len]
            time_delta_seqs: [batch_size, seq_len]
            type_seqs: [batch_size, seq_len]
            sample_times: [batch_size, seq_len, num_samples]

        Returns:
            intensities at sample times: [batch_size, seq_len, num_samples, num_event_types]
        """
        attention_mask = kwargs.get('attention_mask', None)
        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        if attention_mask is None:
            batch_size, seq_len = get_shape_list(time_seqs)
            attention_mask = tf.ones((seq_len, seq_len))
            # only keep the strict upper triangular
            lower_diag_masks = tf.linalg.LinearOperatorLowerTriangular(tf.ones_like(attention_mask)).to_dense()
            attention_mask = tf.where(tf.equal(lower_diag_masks, 0),
                                      attention_mask,
                                      tf.zeros_like(attention_mask))
            attention_mask = tf.tile(attention_mask[None, ...], (batch_size, 1, 1))
            attention_mask = tf.cast(attention_mask, tf.int32)

        # [batch_size, seq_len, num_samples, hidden_size]
        encoder_output = self.compute_states_at_sample_times(time_seqs, type_seqs, attention_mask, sample_times)

        if compute_last_step_only:
            lambdas = self.layer_intensity(encoder_output[:, -1:, :, :])
        else:
            # [batch_size, seq_len, num_samples, num_event_types]
            lambdas = self.layer_intensity(encoder_output)
        return lambdas



import math

import torch
from torch import nn


def attention(query, key, value, mask=None, dropout=None):
    d_k = query.size(-1)
    scores = torch.matmul(query, key.transpose(-2, -1)) / math.sqrt(d_k)
    if mask is not None:
        # small change here -- we use "1" for masked element
        scores = scores.masked_fill(mask > 0, -1e9)
    p_attn = torch.softmax(scores, dim=-1)
    if dropout is not None:
        p_attn = dropout(p_attn)
    return torch.matmul(p_attn, value), p_attn


class MultiHeadAttention(nn.Module):
    def __init__(self, n_head, d_input, d_model, dropout=0.1, output_linear=False):
        super(MultiHeadAttention, self).__init__()
        assert d_model % n_head == 0
        self.n_head = n_head
        self.d_k = d_model // n_head
        self.d_v = self.d_k
        self.d_model = d_model
        self.output_linear = output_linear

        if output_linear:
            self.linears = nn.ModuleList(
                [nn.Linear(d_input, d_model) for _ in range(3)] + [nn.Linear(d_model, d_model), ])
        else:
            self.linears = nn.ModuleList([nn.Linear(d_input, d_model) for _ in range(3)])

        self.dropout = nn.Dropout(p=dropout)

    def forward(self, query, key, value, mask, output_weight=False):
        if mask is not None:
            mask = mask.unsqueeze(1)
        nbatches = query.size(0)

        query, key, value = [
            lin_layer(x).view(nbatches, -1, self.n_head, self.d_k).transpose(1, 2)
            for lin_layer, x in zip(self.linears, (query, key, value))
        ]
        x, attn_weight = attention(query, key, value, mask=mask, dropout=self.dropout)

        x = x.transpose(1, 2).contiguous() \
            .view(nbatches, -1, self.n_head * self.d_k)

        if self.output_linear:
            if output_weight:
                return self.linears[-1](x), attn_weight
            else:
                return self.linears[-1](x)
        else:
            if output_weight:
                return x, attn_weight
            else:
                return x


class SublayerConnection(nn.Module):
    # used for residual connection
    def __init__(self, d_model, dropout):
        super(SublayerConnection, self).__init__()
        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):
        return x + self.dropout(sublayer(self.norm(x)))


class EncoderLayer(nn.Module):
    def __init__(self, d_model, self_attn, feed_forward=None, use_residual=False, dropout=0.1):
        super(EncoderLayer, self).__init__()
        self.self_attn = self_attn
        self.feed_forward = feed_forward
        self.use_residual = use_residual
        if use_residual:
            self.sublayer = nn.ModuleList([SublayerConnection(d_model, dropout) for _ in range(2)])
        self.d_model = d_model

    def forward(self, x, mask):
        if self.use_residual:
            x = self.sublayer[0](x, lambda x: self.self_attn(x, x, x, mask))
            if self.feed_forward is not None:
                return self.sublayer[1](x, self.feed_forward)
            else:
                return x
        else:
            return self.self_attn(x, x, x, mask)


class TimePositionalEncoding(nn.Module):
    """Temporal encoding in THP, ICML 2020
    """

    def __init__(self, d_model, max_len=5000, device='cpu'):
        super().__init__()

        pe = torch.zeros(max_len, d_model, device=device).float()
        position = torch.arange(0, max_len, device=device).float().unsqueeze(1)
        div_term = (torch.arange(0, d_model, 2, device=device).float() * -(math.log(10000.0) / d_model)).exp()

        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)

        # [1, max_len, d_model]
        pe = pe.unsqueeze(0)

        self.register_buffer('pe', pe)

    def forward(self, x):
        """Compute time positional encoding defined in Equation (2) in THP model.

        Args:
            x (tensor): time_seqs, [batch_size, seq_len]

        Returns:
            temporal encoding vector, [batch_size, seq_len, model_dim]

        """
        length = x.size(1)

        return self.pe[:, :length]


class TimeShiftedPositionalEncoding(nn.Module):
    """Time shifted positional encoding in SAHP, ICML 2020
    """

    def __init__(self, d_model, max_len=5000, device='cpu'):
        super().__init__()
        # [max_len, 1]
        position = torch.arange(0, max_len, device=device).float().unsqueeze(1)
        # [model_dim //2 ]
        div_term = (torch.arange(0, d_model, 2, device=device).float() * -(math.log(10000.0) / d_model)).exp()

        self.layer_time_delta = nn.Linear(1, d_model // 2, bias=False)

        self.register_buffer('position', position)
        self.register_buffer('div_term', div_term)

    def forward(self, x, interval):
        """

        Args:
            x: time_seq, [batch_size, seq_len]
            interval: time_delta_seq, [batch_size, seq_len]

        Returns:
            Time shifted positional encoding defined in Equation (8) in SAHP model

        """
        phi = self.layer_time_delta(interval.unsqueeze(-1))
        aa = len(x.size())
        if aa > 1:
            length = x.size(1)
        else:
            length = x.size(0)

        arc = (self.position[:length] * self.div_term).unsqueeze(0)

        pe_sin = torch.sin(arc + phi)
        pe_cos = torch.cos(arc + phi)
        pe = torch.cat([pe_sin, pe_cos], dim=-1)

        return pe


class GELU(nn.Module):
    """GeLu activation function
    """

    def forward(self, x):
        return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))


class Identity(nn.Module):

    def forward(self, inputs):
        return inputs


def activation_layer(act_name):
    """Construct activation layers
    Args:
        act_name: str or nn.Module, name of activation function
    Return:
        act_layer: activation layer

    """
    if isinstance(act_name, str):
        if act_name.lower() == 'sigmoid':
            act_layer = nn.Sigmoid()
        elif act_name.lower() == 'linear':
            act_layer = Identity()
        elif act_name.lower() == 'relu':
            act_layer = nn.ReLU(inplace=True)
        elif act_name.lower() == 'prelu':
            act_layer = nn.PReLU()
        elif act_name.lower() == 'gelu':
            act_layer = GELU()
    elif issubclass(act_name, nn.Module):
        act_layer = act_name()
    else:
        raise NotImplementedError

    return act_layer


class DNN(nn.Module):
    """The Multi Layer Percetron
      Input shape
        - nD tensor with shape: ``(batch_size, ..., input_dim)``.
        The most common situation would be a 2D input with shape ``(batch_size, input_dim)``.
      Output shape
        - nD tensor with shape: ``(batch_size, ..., hidden_size[-1])``.
        For instance, for a 2D input with shape ``(batch_size, input_dim)``,
        the output would have shape ``(batch_size, hidden_size[-1])``.
      Arguments
        - **inputs_dim**: input feature dimension.
        - **hidden_size**:list of positive integer, the layer number and units in each layer.
        - **activation**: Activation function to use.
        - **l2_reg**: float between 0 and 1. L2 regularizer strength applied to the kernel weights matrix.
        - **dropout_rate**: float in [0,1). Fraction of the units to dropout.
        - **use_bn**: bool. Whether use BatchNormalization before activation or not.
        - **seed**: A Python integer to use as random seed.
    """

    def __init__(self, inputs_dim, hidden_size, activation='relu', l2_reg=0, dropout_rate=0, use_bn=False,
                 init_std=0.0001):
        super(DNN, self).__init__()
        self.dropout_rate = dropout_rate
        self.dropout = nn.Dropout(dropout_rate)
        self.l2_reg = l2_reg
        self.use_bn = use_bn
        if len(hidden_size) == 0:
            raise ValueError("hidden_units is empty!!")
        hidden_size = [inputs_dim] + list(hidden_size)

        self.linears = nn.ModuleList(
            [nn.Linear(hidden_size[i], hidden_size[i + 1]) for i in range(len(hidden_size) - 1)])

        if self.use_bn:
            self.bn = nn.ModuleList(
                [nn.BatchNorm1d(hidden_size[i + 1]) for i in range(len(hidden_size) - 1)])

        self.activation_layers = nn.ModuleList(
            [activation_layer(activation) for i in range(len(hidden_size) - 1)])

        for name, tensor in self.linears.named_parameters():
            if 'weight' in name:
                nn.init.normal_(tensor, mean=0, std=init_std)

    def forward(self, inputs):
        deep_input = inputs

        for i in range(len(self.linears)):

            fc = self.linears[i](deep_input)

            if self.use_bn:
                fc = self.bn[i](fc)

            fc = self.activation_layers[i](fc)

            fc = self.dropout(fc)
            deep_input = fc
        return deep_input



from easy_tpp.config_factory.config import Config

from easy_tpp.utils.const import Backend


class TrainerConfig(Config):

    def __init__(self, **kwargs):
        """Initialize the Config class.
        """
        self.seed = kwargs.get('seed', 9899)
        self.gpu = kwargs.get('gpu', -1)
        self.batch_size = kwargs.get('batch_size', 256)
        self.max_epoch = kwargs.get('max_epoch', 10)
        self.shuffle = kwargs.get('shuffle', False)
        self.optimizer = kwargs.get('optimizer', 'adam')
        self.learning_rate = kwargs.get('learning_rate', 1.e-3)
        self.valid_freq = kwargs.get('valid_freq', 1)
        self.use_tfb = kwargs.get('use_tfb', False)
        self.metrics = kwargs.get('metrics', ['acc', 'rmse'])

    def get_yaml_config(self):
        """Return the config in dict (yaml compatible) format.

        Returns:
            dict: config of the trainer specs in dict format.
        """
        return {'seed': self.seed,
                'gpu': self.gpu,
                'batch_size': self.batch_size,
                'max_epoch': self.max_epoch,
                'shuffle': self.shuffle,
                'optimizer': self.optimizer,
                'learning_rate': self.learning_rate,
                'valid_freq': self.valid_freq,
                'use_tfb': self.use_tfb,
                'metrics': self.metrics
                }

    @staticmethod
    def parse_from_yaml_config(yaml_config):
        """Parse from the yaml to generate the config object.

        Args:
            yaml_config (dict): configs from yaml file.

        Returns:
            EasyTPP.TrainerConfig: Config class for trainer specs.
        """
        return TrainerConfig(**yaml_config)

    def copy(self):
        """Copy the config.

        Returns:
            EasyTPP.TrainerConfig: a copy of current config.
        """
        return TrainerConfig(batch_size=self.batch_size,
                             max_epoch=self.max_epoch,
                             shuffle=self.shuffle,
                             optimizer=self.optimizer,
                             learning_rate=self.learning_rate,
                             valid_freq=self.valid_freq,
                             use_tfb=self.use_tfb,
                             metrics=self.metrics
                             )


class ThinningConfig(Config):
    def __init__(self, **kwargs):
        """Initialize the Config class.
        """
        self.num_seq = kwargs.get('num_seq', 10)
        self.num_sample = kwargs.get('num_sample', 1)
        self.num_exp = kwargs.get('num_exp', 500)
        self.look_ahead_time = kwargs.get('look_ahead_time', 10)
        self.patience_counter = kwargs.get('patience_counter', 5)
        self.over_sample_rate = kwargs.get('over_sample_rate', 5)
        self.num_samples_boundary = kwargs.get('num_samples_boundary', 5)
        self.dtime_max = kwargs.get('dtime_max', 5)
        # we pad the sequence at the front only in multi-step generation
        self.num_step_gen = kwargs.get('num_step_gen', 1)

    def get_yaml_config(self):
        """Return the config in dict (yaml compatible) format.

        Returns:
            dict: config of the thinning specs in dict format.
        """
        return {'num_seq': self.num_seq,
                'num_sample': self.num_sample,
                'num_exp': self.num_exp,
                'look_ahead_time': self.look_ahead_time,
                'patience_counter': self.patience_counter,
                'over_sample_rate': self.over_sample_rate,
                'num_samples_boundary': self.num_samples_boundary,
                'dtime_max': self.dtime_max,
                'num_step_gen': self.num_step_gen}

    @staticmethod
    def parse_from_yaml_config(yaml_config):
        """Parse from the yaml to generate the config object.

        Args:
            yaml_config (dict): configs from yaml file.

        Returns:
            EasyTPP.ThinningConfig: Config class for thinning algorithms.
        """
        return ThinningConfig(**yaml_config) if yaml_config is not None else None

    def copy(self):
        """Copy the config.

        Returns:
            EasyTPP.ThinningConfig: a copy of current config.
        """
        return ThinningConfig(num_seq=self.num_seq,
                              num_sample=self.num_sample,
                              num_exp=self.num_exp,
                              look_ahead_time=self.look_ahead_time,
                              patience_counter=self.patience_counter,
                              over_sample_rate=self.over_sample_rate,
                              num_samples_boundary=self.num_samples_boundary,
                              dtime_max=self.dtime_max,
                              num_step_gen=self.num_step_gen)


class BaseConfig(Config):
    def __init__(self, **kwargs):
        """Initialize the Config class.
        """
        self.stage = kwargs.get('stage')
        self.backend = kwargs.get('backend')
        self.dataset_id = kwargs.get('dataset_id')
        self.runner_id = kwargs.get('runner_id')
        self.model_id = kwargs.get('model_id')
        self.exp_id = kwargs.get('exp_id')
        self.base_dir = kwargs.get('base_dir')
        self.specs = kwargs.get('specs', {})
        self.backend = self.set_backend(self.backend)

    @staticmethod
    def set_backend(backend):
        if backend.lower() in ['torch', 'pytorch']:
            return Backend.Torch
        elif backend.lower() in ['tf', 'tensorflow']:
            return Backend.TF
        else:
            raise ValueError(
                f"Backend  should be selected between 'torch or pytorch' and 'tf or tensorflow', "
                f"current value: {backend}"
            )

    def get_yaml_config(self):
        """Return the config in dict (yaml compatible) format.

        Returns:
            dict: config of the base config specs in dict format.
        """
        return {'stage': self.stage,
                'backend': str(self.backend),
                'dataset_id': self.dataset_id,
                'runner_id': self.runner_id,
                'model_id': self.model_id,
                'base_dir': self.base_dir,
                'specs': self.specs}

    @staticmethod
    def parse_from_yaml_config(yaml_config):
        """Parse from the yaml to generate the config object.

        Args:
            yaml_config (dict): configs from yaml file.

        Returns:
            BaseConfig: Config class for trainer specs.
        """
        return BaseConfig(**yaml_config)

    def copy(self):
        """Copy the config.

        Returns:
            BaseConfig: a copy of current config.
        """
        return BaseConfig(stage=self.stage,
                          backend=self.backend,
                          dataset_id=self.dataset_id,
                          runner_id=self.runner_id,
                          model_id=self.model_id,
                          base_dir=self.base_dir,
                          specs=self.specs)


class ModelConfig(Config):
    def __init__(self, **kwargs):
        """Initialize the Config class.
        """
        self.rnn_type = kwargs.get('rnn_type', 'LSTM')
        self.hidden_size = kwargs.get('hidden_size', 32)
        self.time_emb_size = kwargs.get('time_emb_size', 16)
        self.num_layers = kwargs.get('num_layers', 2)
        self.num_heads = kwargs.get('num_heads', 2)
        self.mc_num_sample_per_step = kwargs.get('mc_num_sample_per_step', 20)
        self.sharing_param_layer = kwargs.get('sharing_param_layer', False)
        self.loss_integral_num_sample_per_step = kwargs.get('loss_integral_num_sample_per_step', 20)
        self.dropout_rate = kwargs.get('dropout_rate', 0.0)
        self.use_ln = kwargs.get('use_ln', False)
        self.thinning = ThinningConfig.parse_from_yaml_config(kwargs.get('thinning'))
        self.is_training = kwargs.get('training', False)
        self.num_event_types_pad = kwargs.get('num_event_types_pad', None)
        self.num_event_types = kwargs.get('num_event_types', None)
        self.pad_token_id = kwargs.get('event_pad_index', None)
        self.model_id = kwargs.get('model_id', None)
        self.pretrained_model_dir = kwargs.get('pretrained_model_dir', None)
        self.gpu = kwargs.get('gpu', -1)
        self.model_specs = kwargs.get('model_specs', {})

    def get_yaml_config(self):
        """Return the config in dict (yaml compatible) format.

        Returns:
            dict: config of the model config specs in dict format.
        """
        return {'rnn_type': self.rnn_type,
                'hidden_size': self.hidden_size,
                'time_emb_size': self.time_emb_size,
                'num_layers': self.num_layers,
                'mc_num_sample_per_step': self.mc_num_sample_per_step,
                'sharing_param_layer': self.sharing_param_layer,
                'loss_integral_num_sample_per_step': self.loss_integral_num_sample_per_step,
                'dropout_rate': self.dropout_rate,
                'use_ln': self.use_ln,
                # for some models / cases we may not need to pass thinning config
                # e.g., for intensity-free model
                'thinning': None if self.thinning is None else self.thinning.get_yaml_config(),
                'num_event_types_pad': self.num_event_types_pad,
                'num_event_types': self.num_event_types,
                'event_pad_index': self.pad_token_id,
                'model_id': self.model_id,
                'pretrained_model_dir': self.pretrained_model_dir,
                'gpu': self.gpu,
                'model_specs': self.model_specs}

    @staticmethod
    def parse_from_yaml_config(yaml_config):
        """Parse from the yaml to generate the config object.

        Args:
            yaml_config (dict): configs from yaml file.

        Returns:
            ModelConfig: Config class for trainer specs.
        """
        return ModelConfig(**yaml_config)

    def copy(self):
        """Copy the config.

        Returns:
            ModelConfig: a copy of current config.
        """
        return ModelConfig(rnn_type=self.rnn_type,
                           hidden_size=self.hidden_size,
                           time_emb_size=self.time_emb_size,
                           num_layers=self.num_layers,
                           mc_num_sample_per_step=self.mc_num_sample_per_step,
                           sharing_param_layer=self.sharing_param_layer,
                           loss_integral_num_sample_per_step=self.loss_integral_num_sample_per_step,
                           dropout_rate=self.dropout_rate,
                           use_ln=self.use_ln,
                           thinning=self.thinning,
                           num_event_types_pad=self.num_event_types_pad,
                           num_event_types=self.num_event_types,
                           event_pad_index=self.pad_token_id,
                           pretrained_model_dir=self.pretrained_model_dir,
                           gpu=self.gpu,
                           model_specs=self.model_specs)



from easy_tpp.preprocess.data_loader import TPPDataLoader, EventTokenizer, TPPDataset, get_data_loader

__all__ = ['TPPDataLoader',
           'EventTokenizer',
           'TPPDataset',
           'get_data_loader']



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers

from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


class ContTimeLSTMCell(keras.Model):
    """LSTM Cell in Neural Hawkes Process.
    """

    def __init__(self, hidden_size):
        """Initialize the continuous LSTM Cell

        Args:
            hidden_size (int): size of hidden states.
        """
        super(ContTimeLSTMCell, self).__init__()

        with tf.variable_scope('ContTimeLSTMCell'):
            self.hidden_size = hidden_size
            self.init_dense_layer()

    def init_dense_layer(self):
        """Initialize related dense layers.
        """
        self.layer_input = layers.Dense(self.hidden_size,
                                        activation=tf.nn.sigmoid,
                                        name='layer_input')
        self.layer_forget = layers.Dense(self.hidden_size,
                                         activation=tf.nn.sigmoid,
                                         name='layer_forget')
        self.layer_output = layers.Dense(self.hidden_size,
                                         activation=tf.nn.sigmoid,
                                         name='layer_output')
        self.layer_input_bar = layers.Dense(self.hidden_size,
                                            activation=tf.nn.sigmoid,
                                            name='layer_input_bar')
        self.layer_forget_bar = layers.Dense(self.hidden_size,
                                             activation=tf.nn.sigmoid,
                                             name='layer_forget_bar')

        self.layer_pre_c = layers.Dense(self.hidden_size,
                                        activation=tf.nn.tanh,
                                        name='layer_z')
        self.layer_decay = layers.Dense(self.hidden_size,
                                        activation=tf.nn.softplus,
                                        name='layer_decay')

    def init_state(self, batch_size):
        """Initialize hidden and cell states with zeros.

        Args:
            batch_size (tensor): size of the batch.

        Returns:
            tuple: rnn state, a tuple of three tensors and decay states, a tuple of four tensors.
        """
        zero_dims = tf.stack([batch_size, self.hidden_size])
        rnn_state = (tf.fill(zero_dims, 0.0),
                     tf.fill(zero_dims, 0.0),
                     tf.fill(zero_dims, 0.0))
        decay_state = (tf.fill(zero_dims, 0.0),
                       tf.fill(zero_dims, 0.0),
                       tf.fill(zero_dims, 0.0),
                       tf.fill(zero_dims, 0.0))
        return rnn_state, decay_state

    def call(self, x_t, dtime_t, initial_state):
        """Update the continuous time LSTM cell.

        Args:
            x_t (tensor): [batch_size, hidden_size]
            dtime_t (tensor): [batch_size, 1]
            initial_state (tuple): states initialized in function init_state.

        Returns:
            tuple: updated hidden state and tuple of rnn and decay states.
        """
        # parameter process
        h_t, c_func_t, c_bar_t = initial_state[0]
        input_t = tf.concat([x_t, h_t], axis=-1)

        # update input gate - Equation (5a)
        gate_input = self.layer_input(input_t)

        # update forget gate - Equation (5b)
        gate_forget = self.layer_forget(input_t)

        # update output gate - Equation (5d)
        gate_output = self.layer_output(input_t)

        # update input bar - similar to Equation (5a)
        gate_input_bar = self.layer_input_bar(input_t)

        # update forget bar - similar to Equation (5b)
        gate_forget_bar = self.layer_forget_bar(input_t)

        # update gate decay - Equation (6c)
        gate_decay = self.layer_decay(input_t)

        # update gate z - Equation (5c)
        z_t = self.layer_pre_c(input_t)

        # update cell state to t_i+ - Equation (6a)
        c_t = gate_forget * c_func_t + gate_input * z_t

        # update cell state bar - Equation (6b)
        c_bar_t = gate_forget_bar * c_bar_t + gate_input_bar * z_t

        c_func_t, h_t = ContTimeLSTMCell.decay(c_t, c_bar_t, gate_decay, gate_output, dtime_t)

        rnn_state = (h_t, c_func_t, c_bar_t)
        decay_state = (c_t, c_bar_t, gate_decay, gate_output)
        return h_t, (rnn_state, decay_state)

    @staticmethod
    def decay(cell_i, cell_bar_i, gate_decay, gate_output, dtime):
        """Cell and hidden state decay - Equation (7)

        Args:
            cell_i (tensor): cell state, [batch_size, hidden_size].
            cell_bar_i (tensor): cell bar state, [batch_size, hidden_size].
            gate_decay (tensor): decay state, [batch_size, hidden_size].
            gate_output (tensor): output state, [batch_size, hidden_size].
            dtime (tensor): delta time, , [batch_size, 1].

        Returns:
            tuple: cell state and hidden state.
        """
        c_t = cell_bar_i + (cell_i - cell_bar_i) * tf.math.exp(-gate_decay * dtime)
        h_t = gate_output * tf.tanh(c_t)
        return c_t, h_t

    def dynamic_run(self, seq_type_embed, dtime):
        """Update the continuous time LSTM for all time steps.

        Args:
            seq_type_embed (tensor): [batch_size, seq_len, hidden_size].
            dtime (tensor): [batch_size, seq_len].

        Returns:
            tuple: hidden state, [batch_size, seq_len, hidden_size] and decay state,
            [batch_size, 4, seq_len, hidden_size].
        """

        def move_forward_fn(accumulator, item):
            init_state = accumulator[1]
            x_t = item[0]
            dtime_t = item[1]
            h_t, init_state = self.call(x_t, dtime_t, initial_state=init_state)
            return h_t, init_state

        initial_state = self.init_state(tf.shape(seq_type_embed)[0])
        initial_h_t = initial_state[0][0]

        # Scan(move forward) along times axis
        h_ts, cell_states = tf.scan(move_forward_fn,
                                    (tf.transpose(seq_type_embed, perm=[1, 0, 2]),
                                     tf.transpose(tf.expand_dims(dtime, -1), perm=[1, 0, 2])),
                                    initializer=(initial_h_t, initial_state))

        # Transpose the tensor so that batch_size is in the first dimension
        h_ts = tf.transpose(h_ts, perm=[1, 0, 2])
        decay_states = tf.stack(cell_states[1])
        decay_states = tf.transpose(decay_states, perm=[2, 1, 0, 3])

        return h_ts, decay_states


class NHP(TfBaseModel):

    def __init__(self,
                 model_config):
        """Initialize the model

        Args:
            model_config (dict): config of model specs.
        """
        super(NHP, self).__init__(model_config)

    def build_graph(self):
        """Build up the network
        """
        with tf.variable_scope('NHP'):
            self.build_input_graph()

            # Initialize the rnn cell
            self.rnn_cell = ContTimeLSTMCell(self.hidden_size)

            self.layer_intensity = layers.Dense(self.num_event_types, activation=tf.nn.softplus)

        # Compute the loss
        self.loss, self.num_event = self.loglike_loss()

        # Make predictions
        if self.event_sampler and self.gen_config.num_step_gen == 1:
            self.dtime_predict_one_step, self.type_predict_one_step = \
                self.predict_one_step_at_every_event(self.time_seqs,
                                                     self.time_delta_seqs,
                                                     self.type_seqs)

        if self.event_sampler and self.gen_config.num_step_gen > 1:
            # make generations
            self.dtime_generation, self.type_generation = \
                self.predict_multi_step_since_last_event(self.time_seqs,
                                                         self.time_delta_seqs,
                                                         self.type_seqs,
                                                         num_step=self.gen_config.num_step_gen)

    def forward(self, dtimes_seq, event_seq, len_seq=None):
        """ Move forward through the network """

        # shape - (batch_size, seq_len, hidden_size)
        types_seq_emb = self.layer_type_emb(event_seq)

        h_ts, decay_states = self.rnn_cell.dynamic_run(types_seq_emb,
                                                       dtimes_seq)

        if len_seq is not None:
            # Get last decay state for every seq
            # Find out the index position of last decay states for each sequence
            ind_shape = tf.shape(len_seq)
            ndind = tf.concat([tf.range(ind_shape[0])[:, None], len_seq], axis=-1)

            # shape  (batch_size, 4, hidden_size)
            last_decay_states = tf.gather_nd(decay_states, ndind)
        else:
            last_decay_states = decay_states[:, -1, :, :]

        # h_ts   (batch_size, seq_len, event_num)
        # decay_states (batch_size, seq_len, 4, hidden_size)
        # last_decay_states (batch_size, 4, hidden_size)
        return h_ts, decay_states, last_decay_states

    def loglike_loss(self):
        """Compute the loglike loss.

        Returns:
            tuple: loglike loss and num of events.
        """
        hiddens_ti, decay_states, _ = self.forward(self.time_delta_seqs[:, 1:], self.type_seqs[:, :-1])

        # Lambda(t) right before each event time point
        # lambda_at_event - [batch_size, seq_len = max_len-1, num_event_types]
        # Here we drop the last event because it has no delta_time label (can not decay)
        lambda_at_event = self.layer_intensity(hiddens_ti)

        # interval_t_sample - [batch_size, seq_len = max_len-1, num_mc_sample]
        # for every batch and every event point => do a sampling (num_mc_sampling)
        # the first dtime is zero, so we use time_delta_seq[:, 1:]
        interval_t_sample = self.make_dtime_loss_samples(self.time_delta_seqs[:, 1:])

        # [batch_size, seq_len = max_len - 1, num_mc_sample]
        state_t_sample = self.compute_states_at_sampled_times(decay_states, interval_t_sample)

        # [batch_size, seq_len = max_len - 1, num_mc_sample, event_num]
        lambda_t_sample = self.layer_intensity(state_t_sample)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=self.time_delta_seqs[:, 1:],
                                                                        seq_mask=self.batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=self.type_mask[:, 1:])

        # (num_samples, num_times)
        loss = - tf.reduce_sum(event_ll - non_event_ll)

        return loss, num_events

    def compute_states_at_sampled_times(self, decay_states, sample_dtimes, compute_last_step_only=False):
        """Compute the hidden states at sampled times.

        Args:
            decay_states (tensor): [batch_size, 4, seq_len, hidden_size].
            sample_dtimes (tensor): [batch_size, seq_len, num_samples].
            compute_last_step_only (bool, optional): whether to compute the last step only. Defaults to False.

        Returns:
            tensor: hidden state at each sampled time.
        """

        # update the states given last event
        # cells (batch_size, num_times, hidden_dim)
        cells, cell_bars, decays, outputs = tf.unstack(decay_states, 4, axis=-2)

        if compute_last_step_only:
            _, h_ts = self.rnn_cell.decay(cells[:, -1:, None, :],
                                          cell_bars[:, -1:, None, :],
                                          decays[:, -1:, None, :],
                                          outputs[:, -1:, None, :],
                                          sample_dtimes[:, -1:, :, None])

        else:
            # Use broadcasting to compute the decays at all time steps
            # at all sample points
            # h_ts shape (batch_size, num_times, num_mc_sample, hidden_dim)
            # cells[:, :, None, :]  (batch_size, num_times, 1, hidden_dim)
            _, h_ts = self.rnn_cell.decay(cells[:, :, None, :],
                                          cell_bars[:, :, None, :],
                                          decays[:, :, None, :],
                                          outputs[:, :, None, :],
                                          sample_dtimes[..., None])

        return h_ts

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sampled_dtimes,
                                            **kwargs):
        """Compute the intensity at sampled times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], sequences of timestamps.
            time_delta_seqs (tensor): [batch_size, seq_len], sequences of delta times.
            type_seqs (tensor): [batch_size, seq_len], sequences of event types.
            sampled_dtimes (tensor): [batch_size, seq_len, num_sample], sampled time delta sequence.

        Returns:
            tensor: intensities as sampled_dtimes, [batch_size, seq_len, num_samples, event_num].
        """

        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        hiddens_ti, decay_states, _ = self.forward(time_delta_seqs, type_seqs)

        # [batch_size, seq_len, num_mc_sample, hidden_dim]
        state_t_sample = self.compute_states_at_sampled_times(decay_states, sampled_dtimes, compute_last_step_only)

        # [batch_size, seq_len, num_samples, event_num]
        lambdas = self.layer_intensity(state_t_sample)

        return lambdas



from abc import abstractmethod
from typing import Any

from easy_tpp.utils import save_yaml_config, load_yaml_config, Registrable, logger


class Config(Registrable):

    def save_to_yaml_file(self, fn):
        """Save the config into the yaml file 'fn'.

        Args:
            fn (str): Target filename.

        Returns:
        """
        yaml_config = self.get_yaml_config()
        save_yaml_config(fn, yaml_config)

    @staticmethod
    def build_from_yaml_file(yaml_fn, **kwargs):
        """Load yaml config file from disk.

        Args:
            yaml_fn (str): Path of the yaml config file.

        Returns:
            EasyTPP.Config: Config object corresponding to cls.
        """
        config = load_yaml_config(yaml_fn)
        pipeline_config = config.get('pipeline_config_id')
        config_cls = Config.by_name(pipeline_config.lower())
        logger.critical(f'Load pipeline config class {config_cls.__name__}')
        return config_cls.parse_from_yaml_config(config, **kwargs)

    @abstractmethod
    def get_yaml_config(self):
        """Get the yaml format config from self.

        Returns:
        """
        pass

    @staticmethod
    @abstractmethod
    def parse_from_yaml_config(yaml_config):
        """Parse from the yaml to generate the config object.

        Args:
            yaml_config (dict): configs from yaml file.

        Returns:
            EasyTPP.Config: Config class for data.
        """
        pass

    @abstractmethod
    def copy(self):
        """Get a same and freely modifiable copy of self.

        Returns:
        """
        pass

    def __str__(self):
        """Str representation of the config.

        Returns:
            str: str representation of the dict format of the config.
        """
        return str(self.get_yaml_config())

    def update(self, config):
        """Update the config.

        Args:
            config (dict): config dict.

        Returns:
            EasyTPP.Config: Config class for data.
        """
        logger.critical(f'Update config class {self.__class__.__name__}')
        return self.parse_from_yaml_config(config)

    def pop(self, key: str, default_var: Any):
        """pop out the key-value item from the config.

        Args:
            key (str): key name.
            default_var (Any): default value to pop.

        Returns:
            Any: value to pop.
        """
        return vars(self).pop(key) or default_var

    def get(self, key: str, default_var: Any):
        """Retrieve the key-value item from the config.

        Args:
            key (str): key name.
            default_var (Any): default value to pop.

        Returns:
            Any: value to get.
        """
        return vars(self)[key] or default_var

    def set(self, key: str, var_to_set: Any):
        """Set the key-value item from the config.

        Args:
            key (str): key name.
            var_to_set (Any): default value to pop.

        Returns:
            Any: value to get.
        """
        vars(self)[key] = var_to_set



""" Initialize a Tf model wrapper that feed into Model Runner   """

import tensorflow as tf

from easy_tpp.utils import RunnerPhase
from easy_tpp.utils.tf_utils import set_device, set_optimizer

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


class TfModelWrapper:
    def __init__(self, model, base_config, model_config, trainer_config):
        """A wrapper class for Tensorflow backends.

        Args:
            model (BaseModel): a TPP model.
            base_config (EasyTPP.Config): basic configs.
            model_config (EasyTPP.Config): model spec configs.
            trainer_config (EasyTPP.Config): trainer spec configs.
        """
        self.model = model
        self.base_config = base_config
        self.model_config = model_config
        self.trainer_config = trainer_config
        set_device(self.trainer_config.gpu)

        # init session and build model
        tf.reset_default_graph()
        self.sess = tf.Session()
        self.model.build_graph()
        if self.model_config.is_training:
            # set up optimizer
            optimizer = self.trainer_config.optimizer
            self.learning_rate = self.trainer_config.learning_rate
            self.opt = set_optimizer(optimizer, self.learning_rate)
            self.train_op = self.opt.minimize(self.model.loss)

        # set up tensorboard
        self.use_tfb = self.trainer_config.use_tfb
        self.train_summary_writer, self.valid_summary_writer = None, None
        if self.use_tfb:
            self.train_summary_writer = tf.summary.FileWriter(self.base_config.spec['tfb_train_dir'])
            self.valid_summary_writer = tf.summary.FileWriter(self.base_config.spec['tfb_valid_dir'])

        # init variable and saver
        self.sess.run(tf.global_variables_initializer())
        self.saver = tf.train.Saver()

    def restore(self, ckpt_dir):
        """Load the checkpoint to restore the model.

        Args:
            ckpt_dir (str): path for the checkpoint.
        """
        self.saver.restore(self.sess, ckpt_dir)

    def save(self, ckpt_dir):
        """Save the checkpoint for the model.

        Args:
            ckpt_dir (str): path for the checkpoint.
        """
        self.saver.save(self.sess, ckpt_dir)

    def write_summary(self, epoch, kv_pairs, phase):
        """Write the kv_paris into the tensorboard.

        Args:
            epoch (int): epoch index in the training.
            kv_pairs (dict): metrics dict.
            phase (RunnerPhase): a const that defines the stage of model runner.
        """
        if self.use_tfb:
            summary_writer = None
            if phase == RunnerPhase.TRAIN:
                summary_writer = self.train_summary_writer
            elif phase == RunnerPhase.VALIDATE:
                summary_writer = self.valid_summary_writer
            elif phase == RunnerPhase.PREDICT:
                pass

            metric_summary = tf.Summary()
            if summary_writer is not None:
                for k, v in kv_pairs.items():
                    if k != 'num_events':
                        metric_summary.value.add(tag=k, simple_value=v)
                        summary_writer.add_summary(metric_summary, epoch)

                summary_writer.flush()
        return

    def close_summary(self):
        """Close the tensorboard summary writer.
        """
        if self.train_summary_writer is not None:
            self.train_summary_writer.close()

        if self.valid_summary_writer is not None:
            self.valid_summary_writer.close()
        return

    def run_batch(self, batch, phase):
        """Run one batch.

        Args:
            batch (EasyTPP.BatchEncoding): preprocessed batch data that go into the model.
            phase (RunnerPhase): a const that defines the stage of model runner.

        Returns:
            tuple: for training and validation we return loss, prediction and labels;
            for prediction we return prediction.
        """
        model = self.model
        sess = self.sess

        time_seqs, time_delta_seqs, type_seqs, batch_non_pad_mask, attention_mask, type_mask = batch.data.values()

        # set mode to train
        is_training = (phase == RunnerPhase.TRAIN)

        fd = {
            model.time_seqs: time_seqs,
            model.time_delta_seqs: time_delta_seqs,
            model.type_seqs: type_seqs,
            model.batch_non_pad_mask: batch_non_pad_mask,
            model.attention_mask: attention_mask,
            model.type_mask: type_mask,
            model.is_training: is_training

        }

        # Assume we dont do prediction on train set
        pred_dtime, pred_type = None, None
        label_dtime, label_type = time_delta_seqs[:, 1:], type_seqs[:, 1:]

        mask = batch_non_pad_mask[:, 1:]

        if phase in (RunnerPhase.TRAIN, RunnerPhase.VALIDATE):
            # set mode to train
            if is_training:
                _, loss, num_event = sess.run([self.train_op,
                                               model.loss,
                                               model.num_event],
                                              feed_dict=fd)
            else:
                loss, num_event = sess.run([model.loss,
                                            model.num_event],
                                           feed_dict=fd)

                if self.model.event_sampler:
                    pred_dtime, pred_type = sess.run([model.dtime_predict_one_step,
                                                      model.type_predict_one_step],
                                                     feed_dict=fd)
            return loss, num_event, (pred_dtime, pred_type), (label_dtime, label_type), (mask,)
        else:
            pred_dtime, pred_type = sess.run([model.dtime_generation,
                                              model.type_generation],
                                             feed_dict=fd)
            num_steps = pred_dtime.shape[-1]
            label_dtime = time_delta_seqs[:, -num_steps:]
            label_type = type_seqs[:, -num_steps:]
            return (pred_dtime, pred_type), (label_dtime, label_type)



import os
import random

import numpy as np
import tensorflow as tf

from easy_tpp.utils.import_utils import is_tf_gpu_available

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


def set_seed(seed=1029):
    """Setup random seed.

    Args:
        seed (int, optional): random seed. Defaults to 1029.
    """
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    tf.random.set_random_seed(seed)


def set_device(gpu=-1):
    """Setup the device.

    Args:
        gpu (int, optional): Defaults to -1.
    """
    if gpu >= 0 and is_tf_gpu_available():
        os.environ['CUDA_VISIBLE_DEVICES'] = str(gpu)
    else:
        os.environ['CUDA_VISIBLE_DEVICES'] = ''
    return


def set_optimizer(optimizer, lr):
    """Setup the optimizer.

    Args:
        optimizer (str): name of the optimizer.
        lr (float): learning rate.

    Raises:
        NotImplementedError: if the optimizer's name is wrong or the optimizer is not supported,
        we raise error.

    Returns:
        tf.train.optimzer: tf optimizer.
    """
    optimizer = optimizer.capitalize() + 'Optimizer'
    try:
        optimizer = getattr(tf.train, optimizer)(learning_rate=lr)
    except Exception:
        raise NotImplementedError("optimizer={} is not supported.".format(optimizer))

    return optimizer


def get_shape_list(x):
    """Deal with dynamic shape in tensorflow cleanly.

    Args:
        x (tensor): input tensor.

    Returns:
        list: shape list of the tensor.
    """
    static = x.shape.as_list()
    dynamic = tf.shape(x)
    return [dynamic[i] if s is None else s for i, s in enumerate(static)]


def tensordot(tensor_a, tensor_b):
    """ Tensor dot function. The last dimension of tensor_a and the first dimension of tensor_b must be the same.

    Args:
        tensor_a (tensor): input tensor.
        tensor_b (tensor): input tensor.

    Returns:
        tensor: the result of tensor_a tensor dot tensor_b.
    """
    last_idx_a = len(tensor_a.get_shape().as_list()) - 1
    return tf.tensordot(tensor_a, tensor_b, [[last_idx_a], [0]])


def swap_axes(tensor, axis1, axis2):
    """Interchange two axes of an tensor.
    :param tensor:
    :param axis1: First axis.
    :param axis2: Second axis.
    :return:
    """
    tensor_perm = list(range(len(tensor.shape.as_list())))
    tensor_perm[axis1] = axis2
    tensor_perm[axis2] = axis1

    return tf.transpose(tensor, perm=tensor_perm)


def create_tensor(shape, value):
    """Creates a tensor with all elements set to be the value.

    Args:
        shape (list): the shape of the target tensor to be created.
        value (float): value to fill the tensor.

    Returns:
        tensor: created tensor with target value filled.
    """
    tensor_shape = tf.stack(shape)
    return tf.fill(tensor_shape, value)


def count_model_params():
    """Count the number of params of the model.

    Args:
        model (tf.keras.Model): a torch model.

    Returns:
        int: total num of the parameters.
    """
    return np.sum([np.prod(v.get_shape().as_list()) for v in tf.trainable_variables()])



import math
from typing import Dict

import numpy as np
from torch.utils.data import Dataset, DataLoader

from easy_tpp.preprocess.data_collator import TPPDataCollator
from easy_tpp.preprocess.event_tokenizer import EventTokenizer
from easy_tpp.utils import py_assert, is_tf_available


class TPPDataset(Dataset):
    def __init__(self, data: Dict):
        self.data_dict = data
        self.time_seqs = self.data_dict['time_seqs']
        self.time_delta_seqs = self.data_dict['time_delta_seqs']
        self.type_seqs = self.data_dict['type_seqs']

    def __len__(self):
        """

        Returns: length of the dataset

        """

        py_assert(len(self.time_seqs) == len(self.type_seqs) and len(self.time_delta_seqs) == len(self.type_seqs),
                  ValueError,
                  f"Inconsistent lengths for data! time_seq_len:{len(self.time_seqs)}, event_len: "
                  f"{len(self.type_seqs)}, time_delta_seq_len: {len(self.time_delta_seqs)}")

        return len(self.time_seqs)

    def __getitem__(self, idx):
        """

        Args:
            idx: iteration index

        Returns:
            dict: a dict of time_seqs, time_delta_seqs and type_seqs element

        """
        return dict({'time_seqs': self.time_seqs[idx], 'time_delta_seqs': self.time_delta_seqs[idx],
                     'type_seqs': self.type_seqs[idx]})

    def to_tf_dataset(self, data_collator: TPPDataCollator, **kwargs):
        """Generate a dataset to use in Tensorflow

        Args:
            data_collator (TPPDataCollator): collator to tokenize the event data.

        Raises:
            ImportError: Tensorflow is not installed.

        Returns:
            tf.keras.utils.Sequence: tf Dataset object for TPP data.
        """
        if is_tf_available():
            import tensorflow as tf

            if tf.__version__ >= '2.0':
                tf = tf.compat.v1
                tf.disable_v2_behavior()
        else:
            raise ImportError("Called a Tensorflow-specific function but Tensorflow is not installed.")

        class TfTPPDataset(tf.keras.utils.Sequence):
            def __init__(self, time_seqs, time_delta_seqs, type_seqs, **kwargs):
                """Initialize the class.

                Args:
                    batch_size (int): size of batch.
                    shuffle (bool): whether to shuffle the data in each batch.

                """
                self.time_seqs = time_seqs
                self.time_delta_seqs = time_delta_seqs
                self.type_seqs = type_seqs
                self.data_len = len(self.time_delta_seqs)
                self.batch_size = kwargs.pop('batch_size')
                self.shuffle = kwargs.pop('shuffle', False)
                self.idx = np.arange(self.data_len)
                self.kwargs = kwargs

            def __getitem__(self, index):
                # get batch indexes from shuffled indexes
                batch_idx = self.idx[index * self.batch_size:(index + 1) * self.batch_size]
                batch = dict({'time_seqs': [self.time_seqs[i] for i in batch_idx],
                              'time_delta_seqs': [self.time_delta_seqs[i] for i in batch_idx],
                              'type_seqs': [self.type_seqs[i] for i in batch_idx]})

                batch = data_collator(batch, **self.kwargs)
                return batch

            def __len__(self):
                # Denotes the number of batches per epoch
                return math.ceil(self.data_len / self.batch_size)

            def on_epoch_end(self):
                # Updates indexes after each epoch
                self.idx = np.arange(self.data_len)
                if self.shuffle:
                    np.random.shuffle(self.idx)

        return TfTPPDataset(self.time_seqs, self.time_delta_seqs, self.type_seqs, **kwargs)


def get_data_loader(dataset: TPPDataset, backend: str, tokenizer: EventTokenizer, **kwargs):
    use_torch = backend == 'torch'

    padding = True if tokenizer.padding_strategy is None else tokenizer.padding_strategy
    truncation = False if tokenizer.truncation_strategy is None else tokenizer.truncation_strategy

    if use_torch:
        data_collator = TPPDataCollator(tokenizer=tokenizer,
                                        return_tensors='pt',
                                        max_length=tokenizer.model_max_length,
                                        padding=padding,
                                        truncation=truncation)

        return DataLoader(dataset,
                          collate_fn=data_collator,
                          **kwargs)
    else:
        # we pass to placeholders
        data_collator = TPPDataCollator(tokenizer=tokenizer,
                                        return_tensors='np',
                                        max_length=tokenizer.model_max_length,
                                        padding=padding,
                                        truncation=truncation)

        return dataset.to_tf_dataset(data_collator, **kwargs)



from collections import defaultdict

from .log_utils import default_logger as logger


class Registrable:
    """Any class that inherits from ``Registrable`` gains access to a named registry for its subclasses. To register them, just decorate them with the classmethod ``@BaseClass.register(name)``.

    After which you can call ``BaseClass.list_available()`` to get the keys for the registered subclasses, and ``BaseClass.by_name(name)`` to get the corresponding subclass.

    Note that the registry stores the subclasses themselves; not class instances. In most cases you would then call ``from_params(params)`` on the returned subclass.
    """

    _registry = defaultdict(dict)
    _default_impl = None

    @classmethod
    def register(cls, name, constructor=None, overwrite=False):
        """Register a class under a particular name.
        Args:
            name (str): The name to register the class under.
            constructor (str): optional (default=None)
                The name of the method to use on the class to construct the object.  If this is given,
                we will use this method (which must be a ``classmethod``) instead of the default
                constructor.
            overwrite (bool) : optional (default=False)
                If True, overwrites any existing models registered under ``name``. Else,
                throws an error if a model is already registered under ``name``.

        # Examples
        To use this class, you would typically have a base class that inherits from ``Registrable``:
        ```python
        class Transform(Registrable):
            ...
        ```
        Then, if you want to register a subclass, you decorate it like this:
        ```python
        @Transform.register("shift-transform")
        class ShiftTransform(Transform):
            def __init__(self, param1: int, param2: str):
                ...
        ```
        Registering a class like this will let you instantiate a class from a config file, where you
        give ``"type": "shift-transform"``, and keys corresponding to the parameters of the ``__init__``
        method (note that for this to work, those parameters must have type annotations).
        If you want to have the instantiation from a config file call a method other than the
        constructor, either because you have several different construction paths that could be
        taken for the same object (as we do in ``Transform``) or because you have logic you want to
        happen before you get to the constructor, you can register a specific ``@classmethod`` as the constructor to use.
        """
        registry = Registrable._registry[cls]

        def add_subclass_to_registry(subclass):
            # Add to registry, raise an error if key has already been used.
            if name in registry:
                if overwrite:
                    message = (
                        f"{name} has already been registered as {registry[name][0].__name__}, but "
                        f"overwrite=True, so overwriting with {cls.__name__}"
                    )
                    logger.info(message)
                else:
                    message = (
                        f"Cannot register {name} as {cls.__name__}; "
                        f"name already in use for {registry[name][0].__name__}"
                    )
                    raise RuntimeError(message)
            registry[name] = (subclass, constructor)
            return subclass

        return add_subclass_to_registry

    @classmethod
    def by_name(cls, name):
        """
        Returns a callable function that constructs an argument of the registered class.  Because
        you can register particular functions as constructors for specific names, this isn't
        necessarily the ``__init__`` method of some class.
        """
        logger.debug(f"instantiating registered subclass {name} of {cls}")
        subclass, constructor = cls.resolve_class_name(name)
        if not constructor:
            return subclass
        else:
            return getattr(subclass, constructor)

    @classmethod
    def resolve_class_name(cls, name):
        """
        Returns the subclass that corresponds to the given ``name``, along with the name of the
        method that was registered as a constructor for that ``name``, if any.
        This method also allows ``name`` to be a fully-specified module name, instead of a name that
        was already added to the ``Registry``.  In that case, you cannot use a separate function as
        a constructor (as you need to call ``cls.register()`` in order to tell us what separate
        function to use).
        """
        if name in Registrable._registry[cls]:
            subclass, constructor = Registrable._registry[cls].get(name)
            return subclass, constructor
        else:
            for base_cls, v in Registrable._registry.items():
                if name in v:
                    subclass, constructor = Registrable._registry[base_cls].get(name)
                    return subclass, constructor

        if "." in name:
            # This might be a fully qualified class name, so we'll try importing its "module"
            # and finding it there.
            parts = name.split(".")
            submodule = ".".join(parts[:-1])
            class_name = parts[-1]
            import importlib
            try:
                module = importlib.import_module(submodule)
            except ModuleNotFoundError:
                raise RuntimeError(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to import module {submodule}"
                )

            try:
                subclass = getattr(module, class_name)
                constructor = None
                return subclass, constructor
            except AttributeError:
                raise RuntimeError(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to find class {class_name} in {submodule}"
                )

        else:
            # is not a qualified class name
            raise RuntimeError(
                f"{name} is not a registered name for {cls.__name__}. "
                "You probably need to use the --include-package flag "
                "to load your custom code. Alternatively, you can specify your choices "
                """using fully-qualified paths, e.g. {"model": "my_module.models.MyModel"} """
                "in which case they will be automatically imported correctly."
            )

    @classmethod
    def list_available(cls):
        """List default first if it exists"""
        keys = list(Registrable._registry[cls].keys())
        default = cls._default_impl

        if default is None:
            return keys
        elif default not in keys:
            raise RuntimeError(f"Default implementation {default} is not registered")
        else:
            return [default] + [k for k in keys if k != default]

    @classmethod
    def registry_dict(cls):
        return Registrable._registry[cls]



import tensorflow as tf
from tensorflow.keras import layers

from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


class RMTPP(TfBaseModel):
    """Tensorflow implementation of Recurrent Marked Temporal Point Process, KDD 2016.
    https://www.kdd.org/kdd2016/papers/files/rpp1081-duA.pdf
    """

    def __init__(self, model_config):
        """Initialize the model

        Args:
            model_config (dict): config of model specs.
        """
        super(RMTPP, self).__init__(model_config)

    def build_graph(self):
        """Build up the network
        """
        with tf.variable_scope('RMTPP'):
            self.build_input_graph()

            # Initialize the rnn cell
            self.layer_rnn = layers.SimpleRNN(self.hidden_size,
                                              return_state=False,
                                              return_sequences=True)

            self.layer_temporal_emb = layers.Dense(self.hidden_size)

            self.layer_hidden = layers.Dense(self.num_event_types)

            self.factor_intensity_base = tf.get_variable(name='intensity_base',
                                                         shape=[1, 1, self.num_event_types],
                                                         initializer=tf.keras.initializers.glorot_uniform())
            self.factor_intensity_current_influence = tf.get_variable(name='intensity_current_influence',
                                                                      shape=[1, 1, self.num_event_types],
                                                                      initializer=tf.keras.initializers.glorot_uniform()
                                                                      )

            # Compute the loss
            self.loss, self.num_event = self.loglike_loss()

            # Make predictions
            if self.event_sampler and self.gen_config.num_step_gen == 1:
                self.dtime_predict_one_step, self.type_predict_one_step = \
                    self.predict_one_step_at_every_event(self.time_seqs,
                                                         self.time_delta_seqs,
                                                         self.type_seqs)

            if self.event_sampler and self.gen_config.num_step_gen > 1:
                # make generations
                self.dtime_generation, self.type_generation = \
                    self.predict_multi_step_since_last_event(self.time_seqs,
                                                             self.time_delta_seqs,
                                                             self.type_seqs,
                                                             num_step=self.gen_config.num_step_gen)

    def state_decay(self, states_to_decay, duration_t):
        """Equation (11), which computes the intensity
        """

        # [batch_size, seq_len, num_event_types]
        states_to_decay_ = self.layer_hidden(states_to_decay)

        # [batch_size, seq_len, num_event_types]
        intensity = tf.exp(
            states_to_decay_ + self.factor_intensity_current_influence * duration_t + self.factor_intensity_base)
        return intensity

    def forward(self, time_seqs, time_delta_seqs, type_seqs):
        """Call the model.

        Args:
            time_seqs (tensor): [batch_size, seq_len], timestamp seqs.
            time_delta_seqs (tensor): [batch_size, seq_len], inter-event time seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.

        Returns:
            list: hidden states, [batch_size, seq_len, hidden_dim], states right before the event happens;
                  stacked decay states,  [batch_size, max_seq_length, 4, hidden_dim], states right after
                  the event happens.
        """

        # [batch_size, seq_len, hidden_size]
        type_emb = self.layer_type_emb(type_seqs)

        # [batch_size, seq_len, hidden_size]
        temporal_emb = self.layer_temporal_emb(time_seqs[..., None])

        # [batch_size, seq_len, hidden_size]
        # states right after the event
        decay_states = self.layer_rnn(type_emb + temporal_emb)

        # States decay - Equation (7) in the paper
        # states before the happening of the next event
        h_t = self.state_decay(decay_states, time_delta_seqs[..., None])

        return h_t, decay_states

    def loglike_loss(self):
        """Compute the loglike loss.

        Returns:
            tuple: loglike loss and num of events.
        """

        lambda_at_event, decay_states = self.forward(self.time_seqs[:, :-1], self.time_delta_seqs[:, 1:],
                                                     self.type_seqs[:, :-1])

        # interval_t_sample - [batch_size, num_times=max_len-1, num_mc_sample]
        # for every batch and every event point => do a sampling (num_mc_sampling)
        # the first dtime is zero, so we use time_delta_seq[:, 1:]
        interval_t_sample = self.make_dtime_loss_samples(self.time_delta_seqs[:, 1:])

        # [batch_size, num_times = max_len - 1, num_mc_sample]
        lambda_t_sample = self.compute_states_at_sample_times(decay_states, interval_t_sample)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=self.time_delta_seqs[:, 1:],
                                                                        seq_mask=self.batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=self.type_mask[:, 1:])

        # (num_samples, num_times)
        loss = - tf.reduce_sum(event_ll - non_event_ll)

        return loss, num_events

    def compute_states_at_sample_times(self, decay_states, sample_dtimes):
        """Compute the hidden states at sampled times.

        Args:
            decay_states (tensor): [batch_size, seq_len, hidden_size].
            sample_dtimes (tensor): [batch_size, seq_len, num_samples].

        Returns:
            tensor: hidden state at each sampled time.
        """
        # update the states given last event

        # Use broadcasting to compute the decays at all time steps
        # decay_states[..., None, :]: [batch_size, seq_len, 1, hidden_size]
        # sample_dtimes[..., None]: [batch_size, seq_len, num_mc_sample, 1]
        # h_ts shape (batch_size, num_times, num_mc_sample, hidden_dim)
        h_ts = self.state_decay(decay_states[..., None, :],
                                sample_dtimes[..., None])

        return h_ts

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sampled_dtimes,
                                            **kwargs):
        """Compute the intensity at sampled times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], sequences of timestamps.
            time_delta_seqs (tensor): [batch_size, seq_len], sequences of delta times.
            type_seqs (tensor): [batch_size, seq_len], sequences of event types.
            sampled_dtimes (tensor): [batch_size, seq_len, num_sample], sampled time delta sequence.

        Returns:
            tensor: intensities as sampled_dtimes, [batch_size, seq_len, num_samples, event_num].
        """

        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        _, decay_states = self.forward(time_seqs, time_delta_seqs, type_seqs)

        if compute_last_step_only:
            interval_t_sample = sampled_dtimes[:, -1:, :, None]
            # [batch_size, 1, num_mc_sample, num_event_types]
            sampled_intensities = self.state_decay(decay_states[:, -1:, None, :],
                                                   interval_t_sample)

        else:
            # interval_t_sample - [batch_size, num_times, num_mc_sample, 1]
            interval_t_sample = sampled_dtimes[..., None]
            # Use broadcasting to compute the decays at all time steps
            # at all sample points
            # sampled_intensities shape (batch_size, num_times, num_mc_sample, hidden_dim)
            # decay_states[:, :, None, :]  (batch_size, num_times, 1, hidden_dim)
            sampled_intensities = self.state_decay(decay_states[..., None, :],
                                                   interval_t_sample)

        return sampled_intensities



from easy_tpp.hpo.optuna_hpo import OptunaTuner


@OptunaTuner.register_trial_func(model_id='default', overwrite=False)
def default_trial(trial, **kwargs):
    setting = {
        "trainer_config": {"max_epoch": "suggest_int(40, 100, log=True)",
                           "batch_size": 256,
                           "optimizer": "adam",
                           "learning_rate": "suggest_float(5e-4, 1e-2, log=True)"},
        "model_config": {"hidden_size": "suggest_int(16, 32)"}
    }
    return setting



import tensorflow as tf

from easy_tpp.model.tf_model.tf_baselayer import EncoderLayer, TimeShiftedPositionalEncoding
from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel
from easy_tpp.utils.tf_utils import get_shape_list

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


class SAHP(TfBaseModel):
    """Tensorflow implementation of Self-Attentive Hawkes Process, ICML 2020.
    https://arxiv.org/abs/1907.07561
    """

    def __init__(self, model_config):
        """Initialize the model

        Args:
            model_config (EasyTPP.ModelConfig): config of model specs.
        """
        super(SAHP, self).__init__(model_config)
        self.d_model = model_config.hidden_size
        self.d_time = model_config.time_emb_size
        self.use_norm = model_config.use_ln

        self.n_layers = model_config.num_layers
        self.n_head = model_config.num_heads
        self.dropout = model_config.dropout_rate

    def build_graph(self):
        """Build up the network
        """
        with tf.variable_scope('THP'):
            self.build_input_graph()

            self.layer_position_emb = TimeShiftedPositionalEncoding(self.hidden_size)

            # Equation (12) - (14)
            # Equation (12): mu
            self.mu = tf.get_variable(name='mu',
                                      shape=[self.d_model, self.num_event_types],
                                      initializer=tf.keras.initializers.glorot_uniform())
            # Equation (13): eta
            self.eta = tf.get_variable(name='eta',
                                       shape=[self.d_model, self.num_event_types],
                                       initializer=tf.keras.initializers.glorot_uniform())
            # Equation (14): gamma
            self.gamma = tf.get_variable(name='gamma',
                                         shape=[self.d_model, self.num_event_types],
                                         initializer=tf.keras.initializers.glorot_uniform())

            self.stack_layers = [EncoderLayer(hidden_size=self.d_model,
                                              num_heads=self.n_head,
                                              dropout_rate=self.dropout) for _ in range(self.n_layers)]

            self.loss, self.num_event = self.loglike_loss()

            # Make predictions
            if self.event_sampler and self.gen_config.num_step_gen == 1:
                self.dtime_predict_one_step, self.type_predict_one_step = \
                    self.predict_one_step_at_every_event(self.time_seqs,
                                                         self.time_delta_seqs,
                                                         self.type_seqs)

            if self.event_sampler and self.gen_config.num_step_gen > 1:
                # make generations
                self.dtime_generation, self.type_generation = \
                    self.predict_multi_step_since_last_event(self.time_seqs,
                                                             self.time_delta_seqs,
                                                             self.type_seqs,
                                                             num_step=self.gen_config.num_step_gen)

    def state_decay(self, encode_state, mu, eta, gamma, duration_t):
        """Equation (15), which computes the pre-intensity states

        Args:
            encode_state (tensor): [batch_size, seq_len, hidden_size].
            mu (tensor): [batch_size, seq_len, hidden_size].
            eta (tensor): [batch_size, seq_len, hidden_size].
            gamma (tensor): [batch_size, seq_len, hidden_size].
            duration_t (tensor): [batch_size, seq_len, num_sample].

        Returns:
            tensor: hidden states at event times.
        """

        # [batch_size, hidden_dim]
        # i did not use the exp operation here because it can easily explode!
        states = tf.matmul(encode_state, mu) + (tf.matmul(encode_state, eta) - tf.matmul(encode_state, mu)) * tf.matmul(
            encode_state, gamma) * duration_t

        return states

    def forward(self, time_seqs, time_delta_seqs, event_seqs, attention_mask):
        """Call the model

        Args:
            time_seqs (tensor): [batch_size, seq_len], timestamp seqs.
            time_delta_seqs (tensor): [batch_size, seq_len], inter-event time seqs.
            event_seqs (tensor): [batch_size, seq_len], event type seqs.
            attention_mask (tensor): [batch_size, seq_len, hidden_size], attention masks.

        Returns:
            tensor: hidden states at event times.
        """
        type_embedding = self.layer_type_emb(event_seqs)
        position_embedding = self.layer_position_emb((time_seqs, time_delta_seqs))

        enc_output = type_embedding + position_embedding

        for enc_layer in self.stack_layers:
            enc_output = enc_layer(
                (enc_output,
                 attention_mask))

        # [batch_size, seq_len, hidden_dim]
        return enc_output

    def loglike_loss(self):
        """Compute the loglike loss.

        Returns:
            tuple: loglike loss and num of events.
        """
        # 1. compute event-loglik
        enc_out = self.forward(self.time_seqs[:, 1:],
                               self.time_delta_seqs[:, 1:],
                               self.type_seqs[:, :-1],
                               self.attention_mask[:, 1:, :-1])

        cell_t = self.state_decay(encode_state=enc_out,
                                  mu=self.mu[None, ...],
                                  eta=self.eta[None, ...],
                                  gamma=self.gamma[None, ...],
                                  duration_t=self.time_delta_seqs[:, 1:, None])

        # [batch_size, seq_len, num_event_types]
        lambda_at_event = tf.nn.softplus(cell_t)

        # 2. compute non-event-loglik (using MC sampling to compute integral)
        # 2.1 sample times
        # [batch_size, seq_len, num_sample]
        sample_dtimes = self.make_dtime_loss_samples(self.time_delta_seqs[:, 1:])

        # 2.2 compute intensities at sampled times
        # [batch_size, num_times = max_len - 1, num_sample, event_num]
        state_t_sample = self.compute_states_at_sample_times(encode_state=enc_out,
                                                             sample_dtimes=sample_dtimes)
        lambda_t_sample = tf.nn.softplus(state_t_sample)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=self.time_delta_seqs[:, 1:],
                                                                        seq_mask=self.batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=self.type_mask[:, 1:])

        # return enc_inten to compute accuracy
        loss = - tf.reduce_sum(event_ll - non_event_ll)

        return loss, num_events

    def compute_states_at_sample_times(self,
                                       encode_state,
                                       sample_dtimes):
        """Compute the hidden states at sampled times.

        Args:
            encode_state (tensor): three tensors with each shape [batch_size, seq_len, hidden_size].
            sample_dtimes (tensor): [batch_size, seq_len, num_samples].

        Returns:
            tensor: [batch_size, seq_len, num_samples, hidden_size]， hidden state at each sampled time.
        """

        cell_states = self.state_decay(encode_state[:, :, None, :],
                                       self.mu[None, None, ...],
                                       self.eta[None, None, ...],
                                       self.gamma[None, None, ...],
                                       sample_dtimes[:, :, :, None])

        return cell_states

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sample_dtimes, **kwargs):
        """Compute the intensity at sampled times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], sequences of timestamps.
            time_delta_seqs (tensor): [batch_size, seq_len], sequences of delta times.
            type_seqs (tensor): [batch_size, seq_len], sequences of event types.
            sampled_dtimes (tensor): [batch_size, seq_len, num_sample], sampled time delta sequence.

        Returns:
            tensor: intensities as sampled_dtimes, [batch_size, seq_len, num_samples, event_num].
        """
        attention_mask = kwargs.get('attention_mask', None)
        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        if attention_mask is None:
            batch_size, seq_len = get_shape_list(time_seqs)
            attention_mask = tf.ones((seq_len, seq_len))
            # only keep the strict upper triangular
            lower_diag_masks = tf.linalg.LinearOperatorLowerTriangular(tf.ones_like(attention_mask)).to_dense()
            attention_mask = tf.where(tf.equal(lower_diag_masks, 0),
                                      attention_mask,
                                      tf.zeros_like(attention_mask))
            attention_mask = tf.tile(attention_mask[None, ...], (batch_size, 1, 1))
            attention_mask = tf.cast(attention_mask, tf.int32)

        # [batch_size, seq_len, num_samples]
        enc_out = self.forward(time_seqs, time_delta_seqs, type_seqs, attention_mask)

        # [batch_size, seq_len, num_samples, hidden_size]
        encoder_output = self.compute_states_at_sample_times(enc_out, sample_dtimes)

        if compute_last_step_only:
            lambdas = tf.nn.softplus(encoder_output[:, -1:, :, :])
        else:
            # [batch_size, seq_len, num_samples, num_event_types]
            lambdas = tf.nn.softplus(encoder_output)
        return lambdas



import copy
import os

from easy_tpp.config_factory.config import Config
from easy_tpp.config_factory.data_config import DataConfig
from easy_tpp.config_factory.model_config import TrainerConfig, ModelConfig, BaseConfig
from easy_tpp.utils import create_folder, logger, get_unique_id, get_stage, RunnerPhase, \
    MetricsHelper, DefaultRunnerConfig, py_assert, is_torch_available, is_tf_available, is_tf_gpu_available, \
    is_torch_gpu_available
from easy_tpp.utils.const import Backend


@Config.register('runner_config')
class RunnerConfig(Config):
    def __init__(self, base_config, model_config, data_config, trainer_config):
        """Initialize the Config class.

        Args:
            base_config (EasyTPP.BaseConfig): BaseConfig object.
            model_config (EasyTPP.ModelConfig): ModelConfig object.
            data_config (EasyTPP.DataConfig): DataConfig object.
            trainer_config (EasyTPP.TrainerConfig): TrainerConfig object
        """
        self.data_config = data_config
        self.model_config = model_config
        self.base_config = base_config
        self.trainer_config = trainer_config

        self.ensure_valid_config()
        self.update_config()

        # save the complete config
        save_dir = self.base_config.specs['output_config_dir']
        self.save_to_yaml_file(save_dir)

        logger.info(f'Save the config to {save_dir}')

    def get_yaml_config(self):
        """Return the config in dict (yaml compatible) format.

        Returns:
            dict: config of the runner config in dict format.
        """
        return {'data_config': self.data_config.get_yaml_config(),
                'base_config': self.base_config.get_yaml_config(),
                'model_config': self.model_config.get_yaml_config(),
                'trainer_config': self.trainer_config.get_yaml_config()}

    @staticmethod
    def parse_from_yaml_config(yaml_config, **kwargs):
        """Parse from the yaml to generate the config object.

        Args:
            yaml_config (dict): configs from yaml file.

        Returns:
            RunnerConfig: Config class for trainer specs.
        """
        direct_parse = kwargs.get('direct_parse', False)
        if not direct_parse:
            exp_id = kwargs.get('experiment_id')
            yaml_exp_config = yaml_config[exp_id]
            dataset_id = yaml_exp_config.get('base_config').get('dataset_id')
            if dataset_id is None:
                dataset_id = DefaultRunnerConfig.DEFAULT_DATASET_ID
            try:
                yaml_data_config = yaml_config['data'][dataset_id]
            except KeyError:
                raise RuntimeError('dataset_id={} is not found in config.'.format(dataset_id))

            data_config = DataConfig.parse_from_yaml_config(yaml_data_config)
            # add exp id to base config
            yaml_exp_config.get('base_config').update(exp_id=exp_id)

        else:
            yaml_exp_config = yaml_config
            data_config = DataConfig.parse_from_yaml_config(yaml_exp_config.get('data_config'))

        base_config = BaseConfig.parse_from_yaml_config(yaml_exp_config.get('base_config'))
        model_config = ModelConfig.parse_from_yaml_config(yaml_exp_config.get('model_config'))
        trainer_config = TrainerConfig.parse_from_yaml_config(yaml_exp_config.get('trainer_config'))

        return RunnerConfig(
            data_config=data_config,
            base_config=base_config,
            model_config=model_config,
            trainer_config=trainer_config
        )

    def ensure_valid_config(self):
        """Do some sanity check about the config, to avoid conflicts in settings.
        """

        # during testing we dont do shuffle by default
        self.trainer_config.shuffle = False

        # during testing we dont apply tfb by default
        self.trainer_config.use_tfb = False

        return

    def update_config(self):
        """Updated config dict.
        """
        model_folder_name = get_unique_id()

        log_folder = create_folder(self.base_config.base_dir, model_folder_name)
        model_folder = create_folder(log_folder, 'models')

        self.base_config.specs['log_folder'] = log_folder
        self.base_config.specs['saved_model_dir'] = os.path.join(model_folder, 'saved_model')
        self.base_config.specs['saved_log_dir'] = os.path.join(log_folder, 'log')
        self.base_config.specs['output_config_dir'] = os.path.join(log_folder,
                                                                   f'{self.base_config.exp_id}_output.yaml')

        if self.trainer_config.use_tfb:
            self.base_config.specs['tfb_train_dir'] = create_folder(log_folder, 'tfb_train')
            self.base_config.specs['tfb_valid_dir'] = create_folder(log_folder, 'tfb_valid')

        current_stage = get_stage(self.base_config.stage)
        is_training = current_stage == RunnerPhase.TRAIN
        self.model_config.is_training = is_training
        self.model_config.gpu = self.trainer_config.gpu

        # update the dataset config => model config
        self.model_config.num_event_types_pad = self.data_config.data_specs.num_event_types_pad
        self.model_config.num_event_types = self.data_config.data_specs.num_event_types
        self.model_config.pad_token_id = self.data_config.data_specs.pad_token_id
        self.model_config.max_len = self.data_config.data_specs.max_len

        # update base config => model config
        model_id = self.base_config.model_id
        self.model_config.model_id = model_id

        if self.base_config.model_id == 'ODETPP' and self.base_config.backend == Backend.TF:
            py_assert(self.data_config.data_specs.padding_strategy == 'max_length',
                      ValueError,
                      'For ODETPP in TensorFlow, we must pad all sequence to '
                      'the same length (max len of the sequences)!')

        run = current_stage
        use_torch = self.base_config.backend == Backend.Torch
        device = 'GPU' if self.trainer_config.gpu >= 0 else 'CPU'

        py_assert(is_torch_available() if use_torch else is_tf_available(), ValueError,
                  f'Backend {self.base_config.backend} is not supported in the current environment yet !')

        if use_torch and device == 'GPU':
            py_assert(is_torch_gpu_available(),
                      ValueError,
                      f'Torch cuda is not supported in the current environment yet!')

        if not use_torch and device == 'GPU':
            py_assert(is_tf_gpu_available(),
                      ValueError,
                      f'Tensorflow GPU is not supported in the current environment yet!')

        critical_msg = '{run} model {model_name} using {device} ' \
                       'with {tf_torch} backend'.format(run=run,
                                                        model_name=model_id,
                                                        device=device,
                                                        tf_torch=self.base_config.backend)

        logger.critical(critical_msg)

        return

    def get_metric_functions(self):
        return MetricsHelper.get_metrics_callback_from_names(self.trainer_config.metrics)

    def get_metric_direction(self, metric_name='rmse'):
        return MetricsHelper.get_metric_direction(metric_name)

    def copy(self):
        """Copy the config.

        Returns:
            RunnerConfig: a copy of current config.
        """
        return RunnerConfig(
            base_config=copy.deepcopy(self.base_config),
            model_config=copy.deepcopy(self.model_config),
            data_config=copy.deepcopy(self.data_config),
            trainer_config=copy.deepcopy(self.trainer_config)
        )



from collections import OrderedDict

from easy_tpp.runner.base_runner import Runner
from easy_tpp.utils import RunnerPhase, logger, MetricsHelper, MetricsTracker, concat_element, save_pickle
from easy_tpp.utils.const import Backend


@Runner.register(name='std_tpp')
class TPPRunner(Runner):
    """Standard TPP runner
    """

    def __init__(self, runner_config, unique_model_dir=False, **kwargs):
        super(TPPRunner, self).__init__(runner_config, unique_model_dir, **kwargs)

        self.metrics_tracker = MetricsTracker()
        if self.runner_config.trainer_config.metrics is not None:
            self.metric_functions = self.runner_config.get_metric_functions()

        self._init_model()

        pretrain_dir = self.runner_config.model_config.pretrained_model_dir
        if pretrain_dir is not None:
            self._load_model(pretrain_dir)

    def _init_model(self):
        """Initialize the model.
        """
        self.use_torch = self.runner_config.base_config.backend == Backend.Torch

        if self.use_torch:
            from easy_tpp.utils import set_seed
            from easy_tpp.model.torch_model.torch_basemodel import TorchBaseModel
            from easy_tpp.torch_wrapper import TorchModelWrapper
            from easy_tpp.utils import count_model_params
            set_seed(self.runner_config.trainer_config.seed)

            self.model = TorchBaseModel.generate_model_from_config(model_config=self.runner_config.model_config)
            self.model_wrapper = TorchModelWrapper(self.model,
                                                   self.runner_config.base_config,
                                                   self.runner_config.model_config,
                                                   self.runner_config.trainer_config)
            num_params = count_model_params(self.model)

        else:
            from easy_tpp.utils.tf_utils import set_seed
            from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel
            from easy_tpp.tf_wrapper import TfModelWrapper
            from easy_tpp.utils.tf_utils import count_model_params
            set_seed(self.runner_config.trainer_config.seed)

            self.model = TfBaseModel.generate_model_from_config(model_config=self.runner_config.model_config)
            self.model_wrapper = TfModelWrapper(self.model,
                                                self.runner_config.base_config,
                                                self.runner_config.model_config,
                                                self.runner_config.trainer_config)
            num_params = count_model_params()

        info_msg = f'Num of model parameters {num_params}'
        logger.info(info_msg)

    def _save_model(self, model_dir, **kwargs):
        """Save the model.

        Args:
            model_dir (str): the dir for model to save.
        """
        if model_dir is None:
            model_dir = self.runner_config.base_config.specs['saved_model_dir']
        self.model_wrapper.save(model_dir)
        logger.critical(f'Save model to {model_dir}')
        return

    def _load_model(self, model_dir, **kwargs):
        """Load the model from the dir.

        Args:
            model_dir (str): the dir for model to load.
        """
        self.model_wrapper.restore(model_dir)
        logger.critical(f'Load model from {model_dir}')
        return

    def _train_model(self, train_loader, valid_loader, **kwargs):
        """Train the model.

        Args:
            train_loader (EasyTPP.DataLoader): data loader for the train set.
            valid_loader (EasyTPP.DataLoader): data loader for the valid set.
        """
        test_loader = kwargs.get('test_loader')
        for i in range(self.runner_config.trainer_config.max_epoch):
            train_metrics = self.run_one_epoch(train_loader, RunnerPhase.TRAIN)

            message = f"[ Epoch {i} (train) ]: train " + MetricsHelper.metrics_dict_to_str(train_metrics)
            logger.info(message)

            self.model_wrapper.write_summary(i, train_metrics, RunnerPhase.TRAIN)

            # evaluate model
            if i % self.runner_config.trainer_config.valid_freq == 0:
                valid_metrics = self.run_one_epoch(valid_loader, RunnerPhase.VALIDATE)

                self.model_wrapper.write_summary(i, valid_metrics, RunnerPhase.VALIDATE)

                message = f"[ Epoch {i} (valid) ]:  valid " + MetricsHelper.metrics_dict_to_str(valid_metrics)
                logger.info(message)

                updated = self.metrics_tracker.update_best("loglike", valid_metrics['loglike'], i)

                message_valid = "current best loglike on valid set is {:.4f} (updated at epoch-{})".format(
                    self.metrics_tracker.current_best['loglike'], self.metrics_tracker.episode_best)

                if updated:
                    message_valid += f", best updated at this epoch"
                    self.model_wrapper.save(self.runner_config.base_config.specs['saved_model_dir'])

                if test_loader is not None:
                    test_metrics = self.run_one_epoch(test_loader, RunnerPhase.VALIDATE)

                    message = f"[ Epoch {i} (test) ]: test " + MetricsHelper.metrics_dict_to_str(test_metrics)
                    logger.info(message)

                logger.critical(message_valid)

        self.model_wrapper.close_summary()

        return

    def _evaluate_model(self, data_loader, **kwargs):
        """Evaluate the model on the valid dataset.

        Args:
            data_loader (EasyTPP.DataLoader): data loader for the valid set

        Returns:
            dict: metrics dict.
        """

        eval_metrics = self.run_one_epoch(data_loader, RunnerPhase.VALIDATE)

        self.model_wrapper.write_summary(0, eval_metrics, RunnerPhase.VALIDATE)

        self.model_wrapper.close_summary()

        message = f"Evaluation result: " + MetricsHelper.metrics_dict_to_str(eval_metrics)

        logger.critical(message)

        return eval_metrics

    def _gen_model(self, data_loader, **kwargs):
        """Generation of the TPP, one-step and multi-step are both supported.
        """

        test_result = self.run_one_epoch(data_loader, RunnerPhase.PREDICT)

        # For the moment we save it to a pkl

        message = f'Save the prediction to pickle file pred.pkl'

        logger.critical(message)

        save_pickle('pred.pkl', test_result)

        return

    def run_one_epoch(self, data_loader, phase):
        """Run one complete epoch.

        Args:
            data_loader: data loader object defined in model runner
            phase: enum, [train, dev, test]

        Returns:
            a dict of metrics
        """
        total_loss = 0
        total_num_event = 0
        epoch_label = []
        epoch_pred = []
        epoch_mask = []
        pad_index = self.runner_config.data_config.data_specs.pad_token_id
        metrics_dict = OrderedDict()
        if phase in [RunnerPhase.TRAIN, RunnerPhase.VALIDATE]:
            for batch in data_loader:
                batch_loss, batch_num_event, batch_pred, batch_label, batch_mask = \
                    self.model_wrapper.run_batch(batch, phase=phase)

                total_loss += batch_loss
                total_num_event += batch_num_event
                epoch_pred.append(batch_pred)
                epoch_label.append(batch_label)
                epoch_mask.append(batch_mask)

            avg_loss = total_loss / total_num_event

            metrics_dict.update({'loglike': -avg_loss, 'num_events': total_num_event})

        else:
            for batch in data_loader:
                batch_pred, batch_label = self.model_wrapper.run_batch(batch, phase=phase)
                epoch_pred.append(batch_pred)
                epoch_label.append(batch_label)

        # we need to improve the code here
        # classify batch_output to list
        pred_exists, label_exists = False, False
        if epoch_pred[0][0] is not None:
            epoch_pred = concat_element(epoch_pred, pad_index)
            pred_exists = True
        if len(epoch_label) > 0 and epoch_label[0][0] is not None:
            epoch_label = concat_element(epoch_label, pad_index)
            label_exists = True
            if len(epoch_mask):
                epoch_mask = concat_element(epoch_mask, False)[0]  # retrieve the first element of concat array
                epoch_mask = epoch_mask.astype(bool)

        if pred_exists and label_exists:
            metrics_dict.update(self.metric_functions(epoch_pred, epoch_label, seq_mask=epoch_mask))

        if phase == RunnerPhase.PREDICT:
            metrics_dict.update({'pred': epoch_pred, 'label': epoch_label})

        return metrics_dict



import torch
from torch import nn

from easy_tpp.model.torch_model.torch_baselayer import DNN
from easy_tpp.model.torch_model.torch_basemodel import TorchBaseModel
from easy_tpp.utils import rk4_step_method


def flatten_parameters(model):
    p_shapes = []
    flat_parameters = []
    for p in model.parameters():
        p_shapes.append(p.size())
        flat_parameters.append(p.flatten())
    return torch.cat(flat_parameters)


class NeuralODEAdjoint(torch.autograd.Function):

    def __init__(self, device):
        super(NeuralODEAdjoint, self).__init__()
        self.device = device

    @staticmethod
    def forward(ctx, z_init, delta_t, ode_fn, solver, num_sample_times, *model_parameters):
        """

        Args:
            ctx:
            input: (tensor): [batch_size]
            model:
            solver:
            delta_t (tensor): [batch_size, num_sample_times]

        Returns:

        """

        ctx.ode_fn = ode_fn
        ctx.solver = solver
        ctx.delta_t = delta_t
        ctx.model_parameters = model_parameters
        ctx.num_sample_times = num_sample_times

        total_state = []
        dt_ratio = 1.0 / num_sample_times
        delta_t = delta_t * dt_ratio
        with torch.no_grad():
            state = z_init
            for i in range(num_sample_times):
                # [batch_size, hidden_size]
                state = solver(diff_func=ode_fn, dt=delta_t, z0=state)
                total_state.append(state)

        # [batch_size, num_samples, hidden_size]
        ctx.save_for_backward(state)

        return state

    @staticmethod
    def backward(ctx, grad_z):
        output_state = ctx.saved_tensors[0]  # return a tuple
        ode_fn = ctx.ode_fn
        solver = ctx.solver
        delta_t = ctx.delta_t
        model_parameters = ctx.model_parameters
        num_sample_times = ctx.num_sample_times

        # Dynamics of augmented system to be calculated backwards in time
        def aug_dynamics(aug_states):
            tmp_z = aug_states[0]
            tmp_neg_a = -aug_states[1]

            with torch.set_grad_enabled(True):
                tmp_z = tmp_z.detach().requires_grad_(True)
                func_eval = ode_fn(tmp_z)
                tmp_ds = torch.autograd.grad(
                    (func_eval,), (tmp_z, *model_parameters),
                    grad_outputs=tmp_neg_a,
                    allow_unused=True,
                    retain_graph=True)

            neg_adfdz = tmp_ds[0]
            neg_adfdtheta = [torch.flatten(var) for var in tmp_ds[1:]]

            return [func_eval, neg_adfdz, *neg_adfdtheta]

        dt_ratio = 1.0 / num_sample_times
        delta_t = delta_t * dt_ratio

        with torch.no_grad():
            # Construct back-state for ode solver
            # reshape variable \theta for batch solving
            init_var_grad = [torch.zeros_like(torch.flatten(var)) for var in model_parameters]

            # [z(t_1), a(t_1), \theta]
            z1 = output_state
            a1 = grad_z
            states = [z1, a1, *init_var_grad]

            for i in range(num_sample_times):
                states = solver(aug_dynamics, -delta_t, states)

            grad_z0 = states[1]

            grad_theta = [torch.reshape(torch.mean(var_grad, dim=0), var.shape) for var, var_grad in
                          zip(model_parameters, states[2:])]

        return (grad_z0, None, None, None, None, *grad_theta)


class NeuralODE(nn.Module):
    def __init__(self, model, solver, num_sample_times, device):
        super().__init__()
        self.model = model
        self.solver = solver
        self.params = [w for w in model.parameters()]
        self.num_sample_times = num_sample_times
        self.device = device

    def forward(self, input_state, delta_time):
        """

        Args:
            input_state: [batch_size, hidden_size]
            return_state:

        Returns:

        """
        output_state = NeuralODEAdjoint.apply(input_state,
                                              delta_time,
                                              self.model,
                                              self.solver,
                                              self.num_sample_times,
                                              *self.params)

        # [batch_size, num_sample_times, hidden_size]
        return output_state


class ODETPP(TorchBaseModel):
    """Torch implementation of a TPP with Neural ODE state evolution, which is a simplified version of TPP in
    https://arxiv.org/abs/2011.04583, ICLR 2021

    code reference: https://msurtsukov.github.io/Neural-ODE/;
    https://github.com/liruilong940607/NeuralODE/blob/master/NeuralODE.py

    """

    def __init__(self, model_config):
        """Initialize the model

        Args:
            model_config (EasyTPP.ModelConfig): config of model specs.
        """
        super(ODETPP, self).__init__(model_config)

        self.layer_intensity = nn.Sequential(
            nn.Linear(self.hidden_size, self.num_event_types),
            nn.Softplus())

        self.event_model = DNN(inputs_dim=self.hidden_size,
                               hidden_size=[self.hidden_size])

        self.ode_num_sample_per_step = model_config.model_specs['ode_num_sample_per_step']
        self.time_factor = model_config.model_specs['time_factor']

        self.solver = rk4_step_method

        self.layer_neural_ode = NeuralODE(model=self.event_model,
                                          solver=self.solver,
                                          num_sample_times=self.ode_num_sample_per_step,
                                          device=self.device)

    def forward(self, time_delta_seqs, type_seqs, **kwargs):
        """Call the model.

        Args:
            time_delta_seqs (tensor): [batch_size, seq_len], inter-event time seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.

        Returns:
            tensor: hidden states at event times.

        """
        # [batch_size, seq_len=max_len-1, hidden_size]
        type_seq_emb = self.layer_type_emb(type_seqs)
        time_delta_seqs_ = time_delta_seqs[..., None]

        total_state_at_event_minus = []
        total_state_at_event_plus = []
        last_state = torch.zeros_like(type_seq_emb[:, 0, :], device=self.device)
        for type_emb, dt in zip(torch.unbind(type_seq_emb, dim=-2),
                                torch.unbind(time_delta_seqs_, dim=-2)):
            dt = dt / self.time_factor
            last_state = self.layer_neural_ode(last_state + type_emb, dt)
            total_state_at_event_minus.append(last_state)
            total_state_at_event_plus.append(last_state + type_emb)

        # [batch_size, seq_len, hidden_size]
        state_ti = torch.stack(total_state_at_event_minus, dim=1)

        # [batch_size, seq_len, hidden_size]
        state_to_evolve = torch.stack(total_state_at_event_plus, dim=1)

        return state_ti, state_to_evolve

    def loglike_loss(self, batch):
        """Compute the loglike loss.

        Args:
            batch (list): batch input.

        Returns:
            list: loglike loss, num events.
        """
        time_seqs, time_delta_seqs, type_seqs, batch_non_pad_mask, _, type_mask = batch

        state_ti, state_ti_plus = self.forward(time_delta_seqs[:, 1:], type_seqs[:, :-1])

        # Num of samples in each batch and num of event time point in the sequence
        batch_size, seq_len, _ = state_ti.size()

        # Lambda(t) right before each event time point
        # lambda_at_event - [batch_size, num_times=max_len-1, num_event_types]
        # Here we drop the last event because it has no delta_time label (can not decay)
        lambda_at_event = self.layer_intensity(state_ti)

        # interval_t_sample - [batch_size, num_times=max_len-1, num_mc_sample]
        # for every batch and every event point => do a sampling (num_mc_sampling)
        # the first dtime is zero, so we use time_delta_seq[:, 1:]
        interval_t_sample = self.make_dtime_loss_samples(time_delta_seqs[:, 1:])

        # [batch_size, num_times = max_len - 1, num_mc_sample, hidden_size]
        sample_state_ti = self.compute_states_at_sample_times(state_ti_plus, interval_t_sample)

        # [batch_size, num_times = max_len - 1, num_mc_sample, event_num]
        lambda_t_sample = self.layer_intensity(sample_state_ti)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=time_delta_seqs[:, 1:],
                                                                        seq_mask=batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=type_mask[:, 1:])

        loss = - (event_ll - non_event_ll).sum()

        return loss, num_events

    def compute_states_at_sample_times(self, state_ti_plus, sample_dtimes):
        """Compute the states at sampling times.

        Args:
            state_ti_plus (tensor): [batch_size, seq_len, hidden_size], states right after the events.
            sample_dtimes (tensor): [batch_size, seq_len, num_samples], delta times in sampling.

        Returns:
            tensor: hiddens states at sampling times.
        """

        # Use broadcasting to compute the decays at all time steps
        # at all sample points
        # h_ts shape (batch_size, seq_len, num_samples, hidden_dim)
        state = self.solver(diff_func=self.event_model,
                            dt=sample_dtimes[..., None],  # [batch_size, seq_len, num_samples, 1]
                            z0=state_ti_plus[..., None, :])  # [batch_size, seq_len, 1, hidden_size]

        return state

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sample_dtimes, **kwargs):
        """Compute the intensity at sampled times, not only event times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], times seqs.
            time_delta_seqs (tensor): [batch_size, seq_len], time delta seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.
            sample_dtimes (tensor): [batch_size, seq_len, num_sample], sampled inter-event timestamps.

        Returns:
            tensor: [batch_size, num_times, num_mc_sample, num_event_types],
                    intensity at each timestamp for each event type.
        """

        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        # forward to the last but one event
        state_ti, state_ti_plus = self.forward(time_delta_seqs, type_seqs, **kwargs)

        # Num of samples in each batch and num of event time point in the sequence
        batch_size, seq_len, _ = state_ti.size()

        # [batch_size, num_sample_times, num_mc_sample, hidden_size]
        sample_state_ti = self.compute_states_at_sample_times(state_ti_plus, sample_dtimes)

        if compute_last_step_only:
            # [batch_size, 1, num_mc_sample, num_event_types]
            sampled_intensities = self.layer_intensity(sample_state_ti[:, -1:, :, :])
        else:
            # [batch_size, num_sample_times, num_mc_sample, num_event_types]
            sampled_intensities = self.layer_intensity(sample_state_ti)

        return sampled_intensities



from enum import Enum


class ExplicitEnum(str, Enum):
    """
    Enum with more explicit error message for missing values.
    """

    def __str__(self):
        return str(self.value)

    @classmethod
    def _missing_(cls, value):
        raise ValueError(
            f"{value} is not a valid {cls.__name__}, please select one of {list(cls._value2member_map_.keys())}"
        )


class PaddingStrategy(ExplicitEnum):
    """
    Possible values for the `padding` argument in [`EventTokenizer.__call__`]. Useful for tab-completion in an
    IDE.
    """

    LONGEST = "longest"
    MAX_LENGTH = "max_length"
    DO_NOT_PAD = "do_not_pad"


class TensorType(ExplicitEnum):
    """
    Possible values for the `return_tensors` argument in [`EventTokenizerBase.__call__`]. Useful for
    tab-completion in an IDE.
    """

    PYTORCH = "pt"
    TENSORFLOW = "tf"
    NUMPY = "np"


class RunnerPhase(ExplicitEnum):
    """Model runner phase enum.
    """
    TRAIN = 'train'
    VALIDATE = 'validate'
    PREDICT = 'predict'


class LossFunction(ExplicitEnum):
    """Loss function for neural TPP model.
    """
    LOGLIKE = 'loglike'
    PARTIAL_TIME_LOSS = 'rmse'
    PARTIAL_EVENT_LOSS = 'accuracy'


class LogConst:
    """Format for log handler.
    """
    DEFAULT_FORMAT = '[%(asctime)s] [%(levelname)s] %(message)s'
    DEFAULT_FORMAT_LONG = '%(asctime)s - %(filename)s[pid:%(process)d;line:%(lineno)d:%(funcName)s]' \
                          ' - %(levelname)s: %(message)s'


class PredOutputIndex:
    """Positional index for the output tuple in ModelRunner.
    """
    TimePredIndex = 0
    TypePredIndex = 1


class DefaultRunnerConfig:
    DEFAULT_DATASET_ID = 'conttime'


class TruncationStrategy(ExplicitEnum):
    """
    Possible values for the `truncation` argument in [`EventTokenizer.__call__`]. Useful for tab-completion in
    an IDE.
    """

    LONGEST_FIRST = "longest_first"
    DO_NOT_TRUNCATE = "do_not_truncate"


class Backend(ExplicitEnum):
    """
    Possible values for the `truncation` argument in [`EventTokenizer.__call__`]. Useful for tab-completion in
    an IDE.
    """

    Torch = 'torch'
    TF = 'tensorflow'



import os
import random

import numpy as np
import torch

from easy_tpp.utils.import_utils import is_torch_mps_available


def set_seed(seed=1029):
    """Setup random seed.

    Args:
        seed (int, optional): random seed. Defaults to 1029.
    """
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True


def set_device(gpu=-1):
    """Setup the device.

    Args:
        gpu (int, optional): num of GPU to use. Defaults to -1 (not use GPU, i.e., use CPU).
    """
    if gpu >= 0:
        if torch.cuda.is_available():
            device = torch.device("cuda:" + str(gpu))
        elif is_torch_mps_available():
            device = torch.device("mps")
    else:
        device = torch.device("cpu")
    return device


def set_optimizer(optimizer, params, lr):
    """Setup the optimizer.

    Args:
        optimizer (str): name of the optimizer.
        params (dict): dict of params for the optimizer.
        lr (float): learning rate.

    Raises:
        NotImplementedError: if the optimizer's name is wrong or the optimizer is not supported,
        we raise error.

    Returns:
        torch.optim: torch optimizer.
    """
    if isinstance(optimizer, str):
        if optimizer.lower() == "adam":
            optimizer = "Adam"
    try:
        optimizer = getattr(torch.optim, optimizer)(params, lr=lr)
    except Exception:
        raise NotImplementedError("optimizer={} is not supported.".format(optimizer))
    return optimizer


def count_model_params(model):
    """Count the number of params of the model.

    Args:
        model (torch.nn.Moduel): a torch model.

    Returns:
        int: total num of the parameters.
    """
    return sum(p.numel() for p in model.parameters())



from dataclasses import dataclass
from typing import Union, Optional

from easy_tpp.preprocess.event_tokenizer import EventTokenizer
from easy_tpp.utils import PaddingStrategy, TruncationStrategy


@dataclass
class TPPDataCollator:
    """
    Data collator that will dynamically pad the inputs of event sequences.

    Args:
        tokenizer ([`EventTokenizer`]):
            The tokenizer used for encoding the data.
        padding (`bool`, `str` or [`~utils.PaddingStrategy`], *optional*, defaults to `True`):
            Select a strategy to pad the returned sequences (according to the model's padding side and padding index)
            among:

            - `True` or `'longest'` (default): Pad to the longest sequence in the batch (or no padding if only a single
              sequence is provided).
            - `'max_length'`: Pad to a maximum length specified with the argument `max_length` or to the maximum
              acceptable input length for the model if that argument is not provided.
            - `False` or `'do_not_pad'`: No padding (i.e., can output a batch with sequences of different lengths).
        max_length (`int`, *optional*):
            Maximum length of the returned list and optionally padding length (see above).
        return_tensors (`str`):
            The type of Tensor to return. Allowable values are "np", "pt" and "tf".
    """

    tokenizer: EventTokenizer
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    truncation: Union[bool, str, TruncationStrategy] = False
    return_tensors: str = "pt"

    def __call__(self, features, return_tensors=None):
        if return_tensors is None:
            return_tensors = self.return_tensors

        batch = self.tokenizer.pad(
            features,
            padding=self.padding,
            max_length=self.max_length,
            truncation=self.truncation,
            return_tensors=return_tensors,
        )

        return batch



import tensorflow as tf
from tensorflow.python.keras import layers

from easy_tpp.model.tf_model.tf_baselayer import MultiHeadAttention
from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel
from easy_tpp.utils.tf_utils import get_shape_list, tensordot, swap_axes, create_tensor

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


class ANHN(TfBaseModel):
    """Tensorflow implementation of Attentive Neural Hawkes Network, IJCNN 2021.
        http://arxiv.org/abs/2211.11758
    """

    def __init__(self, model_config):
        """Initialize the model

        Args:
            model_config (dict): config of model specs.
        """
        super(ANHN, self).__init__(model_config)
        self.d_time = model_config['time_emb_size']
        self.use_norm = model_config['use_ln']

        self.n_layers = model_config['num_layers']
        self.n_head = model_config['num_heads']
        self.dropout = model_config['dropout']

    def build_graph(self):
        """ Build up the network """
        with tf.variable_scope('ANHN'):
            self.build_input_graph()

            self.layer_rnn = layers.LSTM(self.hidden_size,
                                         return_state=False,
                                         return_sequences=True,
                                         name='layer_rnn')

            self.lambda_w = tf.get_variable('lambda_w',
                                            shape=[self.hidden_size, self.num_event_types],
                                            dtype=tf.float32,
                                            initializer=tf.glorot_normal_initializer())
            self.lambda_b = tf.get_variable('lambda_b',
                                            shape=[self.num_event_types],
                                            dtype=tf.float32,
                                            initializer=tf.constant_initializer(0.1))

            self.layer_time_delta = layers.Dense(self.hidden_size,
                                                 activation=tf.nn.softplus,
                                                 name='layer_time_delta')
            self.layer_base_intensity = layers.Dense(self.hidden_size,
                                                     activation=tf.nn.sigmoid,
                                                     name='layer_mu')

            self.layer_att = MultiHeadAttention(self.hidden_size,
                                                self.n_head,
                                                self.dropout)

            self.layer_intensity = layers.Dense(self.num_event_types, activation=tf.nn.softplus)

            self.loss, self.num_event = self.loglike_loss()

            # Make predictions
            if self.gen_config and self.gen_config['num_step_gen'] == 1:
                self.dtime_predict_one_step, self.type_predict_one_step = \
                    self.predict_one_step_at_every_event(self.time_seqs,
                                                         self.time_delta_seqs,
                                                         self.type_seqs)

            if self.gen_config and self.gen_config['num_step_gen'] > 1:
                # make generations
                self.dtime_generation, self.type_generation = \
                    self.predict_multi_step_since_last_event(self.time_seqs,
                                                             self.time_delta_seqs,
                                                             self.type_seqs,
                                                             num_step=self.gen_config['num_step_gen'])

    def forward(self, dtime_seqs, type_seqs, attention_mask):
        """Call the model.

        Args:
            dtime_seqs (tensor): [batch_size, seq_len].
            type_seqs (tensor): [batch_size, seq_len].
            attention_mask (tensor): [batch_size, seq_len, hidden_size].

        Returns:
            list: hidden states, [batch_size, seq_len, hidden_size], states right before the event happens;
                  stacked decay states,  [batch_size, max_seq_length, 4, hidden_dim], states right after
                  the event happens.
        """

        # [batch_size, seq_len, hidden_size]
        event_emb = self.layer_type_emb(type_seqs)

        # [batch_size, seq_len, hidden_size]
        rnn_output = self.layer_rnn(event_emb)

        # [batch_size, seq_len, hidden_size]
        # mu in Equation (3)
        intensity_base = self.layer_base_intensity(rnn_output)

        # [batch_size, num_head, seq_len, seq_len]
        _, att_weight = self.layer_att(rnn_output, attention_mask=attention_mask, output_weight=True)

        # reformat the tensor
        # [batch_size, seq_len, seq_len, 1]
        att_weight = tf.reduce_mean(att_weight, axis=1)[..., None]

        # At each step, alpha and delta reply on all previous event embeddings because there is a cumsum in Equation
        # (3), therefore the alpha and beta have shape [batch_size, seq_len, seq_len, hidden_size] when performing
        # matrix operations.
        # [batch_size, seq_len, seq_len, hidden_size]
        # alpha in Equation (3)
        intensity_alpha = att_weight * rnn_output[:, None, :, :]

        # compute delta
        max_len = get_shape_list(event_emb)[1]

        # [batch_size, seq_len, seq_len, hidden_dim]
        left = tf.tile(rnn_output[:, None, :, :], [1, max_len, 1, 1])
        right = tf.tile(rnn_output[:, :, None, :], [1, 1, max_len, 1])

        # [batch_size, seq_len, seq_len, hidden_dim * 2]
        cur_prev_concat = tf.concat([left, right], axis=-1)
        # [batch_size, seq_len, seq_len, hidden_dim]
        intensity_delta = self.layer_time_delta(cur_prev_concat)

        # compute time elapse
        # [batch_size, seq_len, seq_len, 1]
        base_dtime, target_cumsum_dtime = self.compute_cumsum_dtime(dtime_seqs)

        # [batch_size, max_len, hidden_size]
        imply_lambdas = self.compute_states_at_event_times(intensity_base,
                                                           intensity_alpha,
                                                           intensity_delta,
                                                           target_cumsum_dtime)

        return imply_lambdas, (intensity_base, intensity_alpha, intensity_delta), (base_dtime, target_cumsum_dtime)

    def loglike_loss(self):
        """Compute the loglike loss.

        Args:
            batch (list): batch input.

        Returns:
            tuple: loglikelihood loss and num of events.
        """

        # 1. compute event-loglik
        imply_lambdas, (intensity_base, intensity_alpha, intensity_delta), (base_dtime, target_cumsum_dtime) \
            = self.forward(self.time_delta_seqs[:, 1:],
                           self.type_seqs[:, :-1],
                           self.attention_mask[:, 1:, :-1])
        lambda_at_event = self.layer_intensity(imply_lambdas)

        # 2. compute non-event-loglik (using MC sampling to compute integral)
        # 2.1 sample times
        # [batch_size, seq_len, num_sample]
        sample_dtimes = self.make_dtime_loss_samples(self.time_delta_seqs[:, 1:])

        # [batch_size, num_times = max_len - 1, num_sample, event_num]
        state_t_sample = self.compute_states_at_sample_times(intensity_base, intensity_alpha, intensity_delta,
                                                             base_dtime, sample_dtimes)
        lambda_t_sample = self.layer_intensity(state_t_sample)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=self.time_delta_seqs[:, 1:],
                                                                        seq_mask=self.batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=self.type_mask[:, 1:])

        # return enc_inten to compute accuracy
        loss = - tf.reduce_sum(event_ll - non_event_ll)

        return loss, num_events

    def compute_cumsum_dtime(self, dtime_seqs):
        """Compute cumulative delta times.

        Args:
            dtime_seqs (tensor): [batch_size, seq_len].

        Returns:
            tensor: [batch_size, seq_len].
        """
        # [batch_size, seq_len, num_sample]
        #  dtime_seqs here = dtime_seqs[:, 1:]
        # [dt_1, dt_2, dt_3] => [dt_1 + dt_2, dt_2, 0]
        cum_dtimes = tf.cumsum(dtime_seqs, axis=1, reverse=True, exclusive=True)

        # [batch_size, seq_len, seq_len, 1] (lower triangular: positive, upper: negative, diagonal: zero)
        base_elapses = tf.expand_dims(cum_dtimes[:, None, :] - cum_dtimes[:, :, None], axis=-1)

        # [batch_size, seq_len, seq_len， 1]
        target_cumsum = base_elapses + dtime_seqs[:, :, None, None]

        return base_elapses, target_cumsum

    def compute_states_at_event_times(self, intensity_base, intensity_alpha, intensity_delta, cumsum_dtimes):
        """Compute implied lambda based on Equation (3).

        Args:
            intensity_base (tensor): [batch_size, seq_len, (num_sample), hidden_size]
            intensity_alpha (tensor): [batch_size, seq_len, seq_len, (num_sample), hidden_size]
            intensity_delta (tensor): [batch_size, seq_len, seq_len, (num_sample), hidden_size]
            cumsum_dtimes: [batch_size, seq_len, (num_sample), 1]

        Returns:
            hidden states at all cumsum_dtimes: [batch_size, seq_len, num_samples, hidden_size]

        """
        # to avoid nan calculated by exp after (nan * 0 = nan)
        elapse = tf.abs(cumsum_dtimes)

        # [batch_size, seq_len, hidden_dim]
        cumsum_term = tf.reduce_sum(intensity_alpha * tf.exp(-intensity_delta * elapse), axis=-2)
        # [batch_size, seq_len, hidden_dim]
        imply_lambdas = intensity_base + cumsum_term

        return imply_lambdas

    def compute_states_at_sample_times(self, intensity_base, intensity_alpha, intensity_delta, base_dtime,
                                       sample_dtimes):
        """Compute the hidden states at sampled times.

        Args:
            intensity_base (tensor): [batch_size, seq_len, hidden_size].
            intensity_alpha (tensor): [batch_size, seq_len, seq_len, hidden_size].
            intensity_delta (tensor): [batch_size, seq_len, seq_len, hidden_size].
            base_dtime (tensor): [batch_size, seq_len, seq_len, hidden_size].
            sample_dtimes (tensor): [batch_size, seq_len, num_samples].

        Returns:
            tensor: hidden state at each sampled time, [batch_size, seq_len, num_sample, hidden_size].
        """
        # [max_len, batch_size, n_sample, hidden_dim]
        mus_trans = tf.transpose(intensity_base, perm=[1, 0, 2])[:, :, None]

        # [max_len, batch_size, n_sample, max_len, hidden_dim]
        alphas_trans = tf.transpose(intensity_alpha, perm=[1, 0, 2, 3])[:, :, None]
        deltas_trans = tf.transpose(intensity_delta, perm=[1, 0, 2, 3])[:, :, None]

        base_elapses_trans = tf.transpose(base_dtime, perm=[1, 0, 2, 3])[:, :, None]

        batch_size, num_sample_per_step, _ = get_shape_list(sample_dtimes)

        state_scan_initializer = create_tensor([batch_size,
                                                num_sample_per_step,
                                                self.hidden_size],
                                               0.0)

        # [seq_len, batch_size, num_sample, hidden_size]
        states_samples = tf.scan(fn=self.get_compute_lambda_forward_fn(),
                                 elems=[
                                     mus_trans,
                                     alphas_trans,
                                     deltas_trans,
                                     base_elapses_trans,
                                     swap_axes(sample_dtimes[:, :, :, None, None], 0, 1)],
                                 initializer=state_scan_initializer)

        # [batch_size, seq_len, num_sample, hidden_size]
        states_samples = swap_axes(states_samples, 1, 0)
        return states_samples

    def get_compute_lambda_forward_fn(self):
        """Compute the lambda using scan function.

        Returns:
            function: a forward function used in tf.scan.
        """
        compute_states_fn = self.compute_states_at_event_times

        def forward_fn(acc, item):
            mu, alpha, delta, elapse, elapse_bias = item
            return compute_states_fn(mu, alpha, delta, elapse + elapse_bias)

        return forward_fn

    def layer_intensity(self, hidden_states):
        """Compute the intensity based on the hidden states.

        Args:
            hidden_states (tensor): [batch_size, seq_len, hidden_size].

        Returns:
            tensor: [batch_size, seq_len, num_event_type_no_pad].
        """
        return tf.nn.softplus(tensordot(hidden_states, self.lambda_w) + self.lambda_b)

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sample_dtimes, **kwargs):
        """Compute the intensity at sampled times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], sequences of timestamps.
            time_delta_seqs (tensor): [batch_size, seq_len], sequences of delta times.
            type_seqs (tensor): [batch_size, seq_len], sequences of event types.
            sampled_dtimes (tensor): [batch_size, seq_len, num_sample], sampled time delta sequence.

        Returns:
            tensor: intensities as sampled_dtimes, [batch_size, seq_len, num_samples, event_num].
        """
        attention_mask = kwargs.get('attention_mask', None)
        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        if attention_mask is None:
            batch_size, seq_len = get_shape_list(time_seqs)
            attention_mask = tf.ones((seq_len, seq_len))
            # only keep the strict upper triangular
            lower_diag_masks = tf.linalg.LinearOperatorLowerTriangular(tf.ones_like(attention_mask)).to_dense()
            attention_mask = tf.where(tf.equal(lower_diag_masks, 0),
                                      attention_mask,
                                      tf.zeros_like(attention_mask))
            attention_mask = tf.tile(attention_mask[None, ...], (batch_size, 1, 1))
            attention_mask = tf.cast(attention_mask, tf.int32)

        # [batch_size, seq_len, num_samples]
        imply_lambdas, (intensity_base, intensity_alpha, intensity_delta), (base_dtime, target_cumsum_dtime) \
            = self.forward(time_delta_seqs, type_seqs, attention_mask)

        # [batch_size, seq_len, num_samples, hidden_size]
        encoder_output = self.compute_states_at_sample_times(intensity_base, intensity_alpha, intensity_delta,
                                                             base_dtime, sample_dtimes)

        if compute_last_step_only:
            lambdas = self.layer_intensity(encoder_output[:, -1:, :, :])
        else:
            # [batch_size, seq_len, num_samples, num_event_types]
            lambdas = self.layer_intensity(encoder_output)
        return lambdas
